import numpy as np
from . import timing
from . import parallel



def magnetic_response_map(h,nk=20,nq=20,j=[0.1,0.,0.],r=[0,0,1],
          kp=None,qs=None):
  """Generate a magnetic susceptibility map"""
  h0 = h.copy() # copy Hamiltonian
# function with the energy
  energy_qvector = energy_qvector_generator(h,j=j,r=r,kp=kp,nk=nk) 
  f = open("SUSCEPTIBILITY.OUT","w")
  k2K = h.geometry.get_k2K_generator() # get the function
  # loop over qvectors
  if qs is None: # not provided
    xs = np.linspace(-2.0,2.0,nq) # kx
    ys = np.linspace(-2.0,2.0,nq) # ky
    qs = [] # initialize qvectors
    for x in xs:
      for y in ys:
        qs.append(np.array([x,y,0.])) # store
  else: pass # qs provided on input
  est = timing.Testimator(maxite=len(qs)) # initialize estimator
  def fun(q0):
    q = k2K(q0) # put in reciprocal coordinates
    return energy_qvector(q=q)
  es = parallel.pcall(fun,qs)
  for (e,q0) in zip(es,qs): # loop
#    est.iterate()
#    e = energy_qvector(q=q)  # energy
    f.write(str(q0[0])+"  "+str(q0[1])+"  "+str(e)+"\n")
    f.flush()
  f.close()
  print("writen SUSCEPTIBILITY.OUT")




def energy_qvector_generator(h,j=[1.0,0.0,0.0],r=[0,0,1],kp=None,nk=10):
  """Return a function that calculates the total energy
  in a specific qvector"""
  h0 = h.copy() # copy Hamiltonian
  def energy_qvector(random=False,q=[0.,0.,0.]):
    h = h0.copy()
    h.generate_spin_spiral(vector=r,qspiral=2.*q)
    h.add_zeeman(j)
    return h.total_energy(nk=nk,kp=kp)
  return energy_qvector





import numpy as np

def reciprocal_evaluation(h,f,nk=50,nsuper=1,reciprocal=True):
    """Compute a reciprocal map"""
    if h.dimensionality!=2: raise  # continue if two dimensional
    kxs = np.linspace(-nsuper,nsuper,nk,endpoint=True)  # generate kx
    kys = np.linspace(-nsuper,nsuper,nk,endpoint=True)  # generate ky
    if reciprocal: fR = h.geometry.get_k2K_generator() # get matrix
    else:  fR = lambda x: x # get identity
    # setup the operator
    rs = [] # empty list
    for x in kxs:
      for y in kys:
        rs.append([x,y,0.]) # store
    def getf(r): 
        return f(fR(r)) # function to compute
    rs = np.array(rs) # transform into array
    from . import parallel
    kx = rs[:,0] # x coordinate
    ky = rs[:,1] # y coordinate
    out = [getf(r) for r in rs]
#    out = parallel.pcall(getf,rs) # compute all
    out = np.array(out) # transform into an array
    return kx,ky,out




def sc_kmap(h,fh,**kwargs):
    """Compute the reciprocal space map of spin singlet SC"""
    h = h.copy()
    h = fh(h)
    hk = h.get_hk_gen() # get Bloch Hamiltonian generator
    def f(k):
        m = hk(k)
        m = m@m
        return np.trace(m).real # return the trace
    kx,ky,out = reciprocal_evaluation(h,f,**kwargs)
    np.savetxt("RECIPROCAL_MAP.OUT",np.array([kx,ky,out]).T)

def singlet_map(h,**kwargs):
    from .sctk.extract import get_singlet_hamiltonian as fh
    return sc_kmap(h,fh,**kwargs)


def triplet_map(h,**kwargs):
    from .sctk.extract import get_triplet_hamiltonian as fh
    return sc_kmap(h,fh,**kwargs)




import numpy as np
from scipy.sparse import csc_matrix


def hartree(h,v=0.0,**kwargs):
    """Add a crystal field to a Hamiltonian"""
    if v==0.0: return
    m = cf_potential(h.geometry,vc=v,**kwargs) # get array
    m = m - np.min(m) # shift to zero
    if np.max(np.abs(m))>1e-7: m = m/np.max(np.abs(m)) # normalize
    m = m - np.mean(m) # remove average
    nat = len(h.geometry.r) # number of atoms
    ind = range(nat) # indexes
    mat = csc_matrix((m,(ind,ind)),shape=(nat,nat),dtype=np.complex)
    h.intra = h.intra + h.spinless2full(mat) # add contribution


def hartree_onsite(g,**kwargs):
    return cf_potential(g,**kwargs,mode="full")



def cf_potential(g,rcut=6.0,vc=0.0,mode="full"):
    """Return an array with the Hartree terms"""
# create fucntions to compute effective distance
    if mode=="full": # full mode
        def getd(dr,dx,dy,dz): return dr
    elif mode=="stacking": # stacking mode
        def getd(dr,dx,dy,dz):
            dr[np.abs(dz)<1e-3] = 1e10
            return dr
    else: raise # not implemented
    g = g.copy() # copy geometry
    interactions = [] # empty list
    nat = len(g.r) # number of atoms
    mout = np.zeros(nat) # initialize matrix
    if g.dimensionality>0:
      lat = np.sqrt(g.a1.dot(g.a1)) # size of the unit cell
      g.ncells = int(2*rcut/lat)+2 # number of unit cells to consider
    ri = g.r # positions
    for d in g.neighbor_directions(): # loop
        rj = np.array(g.replicas(d)) # positions
        for i in range(nat):
            dx = rj[:,0] - ri[i,0]
            dy = rj[:,1] - ri[i,1]
            dz = rj[:,2] - ri[i,2]
            dr = dx*dx + dy*dy + dz*dz
            dr = np.sqrt(dr) # square root
            dr = getd(dr,dx,dy,dz) # effective distance
            dr[dr<1e-4] = 1e10
#            v = vc/dr # Coulomb interaction
            v = vc*np.exp(-dr/rcut) # quench interaction
            #v = v*1./(dr/rcut) # quench interaction
            v[dr<1e-4] = 0.0
            v[dr>rcut] = 0.0
            mout[i] += np.sum(v) # store contribution
    return mout # return



def bulkedge_function(g,vbulk=0.0,vedge=1.0,rcut=20.0,sharpness=None,
        ebcut=.5):
    """Return a function that distinguishes between bulk and edge"""
    m = hartree_onsite(g,vc=1.0,rcut=rcut) # compute the array
    m = m - np.min(m) # shift
    m = m/np.max(m) # normalize
    if not sharpness is None:  
        m = (np.tanh(sharpness*(m-ebcut)) + 1.)/2.
    m = vedge + (vbulk-vedge)*m # scale
    from .geometry import array2function
    f = array2function(g,m) 
    return f # return function







import numpy as np


def radial_vector_decay(**kwargs):
    from .profiles import radial_decay
    f = radial_decay(**kwargs) # function for the radial decay
    def fs(r,dr):
        r0 = np.sqrt(r.dot(r))
        dr0 = np.sqrt(dr.dot(dr))
        if r0<1e-4 or dr0<1e-4: return 1.0
        ur0,udr0 = r/r0,dr/dr0
        udr = np.abs(ur0.dot(udr0))
        return 1.0*(1.-udr) + f(r)*udr # return
    return fs



from __future__ import print_function,division
import numpy as np
import scipy.linalg as lg
from . import sculpt
from numba import jit



def non_orthogonal_supercell(gin,m,ncheck=2,mode="fill",reducef=lambda x: x):
  """Generate a non orthogonal supercell based on a tranformation
  matrix of the unit vectors, pretty much as VESTA does"""
  # workaround
  g = gin.copy()
  if g.dimensionality==0: return
  if g.dimensionality==1:
    g.a2 = np.array([0.,np.max(np.abs(gin.y))*2.+1.,0.])
    g.a3 = np.array([0.,0.,np.max(np.abs(gin.z))*2.+1.])
  if g.dimensionality==2:
    dz = np.max(np.abs(gin.z))*2.+1.
    g.a3 = np.array([0.,0.,dz])
  a1,a2,a3 = g.a1,g.a2,g.a3 # cell vectors
  go = g.copy() # output unit cell
  # new cell vectors
  go.a1 = m[0][0]*a1 + m[0][1]*a2 + m[0][2]*a3
  go.a2 = m[1][0]*a1 + m[1][1]*a2 + m[1][2]*a3
  go.a3 = m[2][0]*a1 + m[2][1]*a2 + m[2][2]*a3
  # calculate old and new volume
  vold = a1.dot(np.cross(a2,a3))  
  vnew = go.a1.dot(np.cross(go.a2,go.a3))  
  if abs(vnew)<0.0001: 
    print("No volume",vnew,"\n",a1,"\n",a2,"\n",a3)
    raise
  c = vnew/vold
  c = int(round(abs(c)))
  # now create replicas until there as c times as many atoms in the
  # unit cell
  if mode=="fill": # look for atoms until everything is filled
    rs = []
    k2K = go.get_k2K().I # matrix transformation
    R = np.matrix([go.a1,go.a2,go.a3]).T # transformation matrix
    L = R.I # inverse matrix
    d0 = -np.random.random()*0.1 # accuracy
    d0 = -0.122132112 # some random number
    d1 = 1.0 + d0 # accuracy
    from .geometry import neighbor_cells
# get as many cells as necessary
    cneigh = reducef(c) # cells to generate given the volume increase c
    cneigh = int(round(cneigh)) # integer
    inds = neighbor_cells(cneigh,dim=g.dimensionality) 
    sl = [] # list for the sublattice
    for (i,j,k) in inds: # loop
          ii = 0 # start count
          for ir in range(len(g.r)): # loop over positions
            ri = g.r[ir] # get the position
            store = False
            rj = ri + i*g.a1 + j*g.a2 + k*g.a3 # new position
            rn = L*np.matrix(rj).T  # transform
            rn = np.array([rn.T[0,ii] for ii in range(3)]) # convert to array
            n1,n2,n3 = rn[0],rn[1],rn[2]
            if g.dimensionality==3 and d0<n1<d1 and d0<n2<d1 and d0<n3<d1:
                store = True
            if g.dimensionality==2 and d0<n1<d1 and d0<n2<d1: 
                store = True
            if g.dimensionality==1 and d0<n1<d1: 
                store = True
            if store:
                rs.append(rj)
                if g.has_sublattice: sl.append(g.sublattice[ir])
#          if len(rs)==len(g.r)*c:
#            print("All the atoms found")
#            break
#    print(rs)
    go.r = np.array(rs) # store
    if go.has_sublattice: go.sublattice = sl # store sublattice
    if len(rs)!=len(g.r)*c: 
      print("Not all the atoms have been found")
      print("New atoms",len(rs))
      print("Expected atoms",len(g.r)*c)
      print("Volume of the cell increase",c)
      raise
  elif mode=="brute":
    if g.dimensionality==1:
      rs3 = replicate3d(g.r,g.a1,g.a2,g.a3,c,1,1) # new positions
    if g.dimensionality==2:
      rs3 = replicate3d(g.r,g.a1,g.a2,g.a3,c,c,1) # new positions
    if g.dimensionality==3:
      rs3 = replicate3d(g.r,g.a1,g.a2,g.a3,c,c,c) # new positions
    while True: # infinite loop, stop when scf reached
      rs1 = np.array(rs3) # store the first iteration
#      print(rs1)
      for i in range(-ncheck,ncheck+1):
        for j in range(-ncheck,ncheck+1):
          for k in range(-ncheck,ncheck+1):
            if i==0 and j==0 and k==0: continue
            rs2 = [ri + i*go.a1 + j*go.a2 + k*go.a3 for ri in rs1] # shift by this vector
            rs1 = return_unique(rs1,rs2) # return the unique positions
#            print(len(rs1),i,j,k)
      if len(rs1)==len(rs3): break
      rs3 = np.array(rs1) 
    go.r = np.array(rs1) # store new positions
    if go.has_sublattice: go.get_sublattice()
  go.r2xyz() # update coordinates
  go.center()
  go.get_fractional()
  return go # return new geometry
  

# from numba import jit


def replicate3d(rs,a1,a2,a3,n1,n2,n3):
  ro = np.zeros((n1*n2*n3*nc,3)) # allocate output array
  return replicate3d_jit(rs,a1,a2,a3,n1,n2,n3,ro) # compute

@jit(nopython=True)
def replicate3d_jit(rs,a1,a2,a3,n1,n2,n3,ro):
  """Function to make a three dimensional supercell"""
  nc = len(rs)
  ik = 0
  for i in range(n1):
    for j in range(n2):
      for l in range(n3):
        for k in range(nc):
          ro[ik] = a1*i + a2*j + a3*l + rs[k] # store position
          ik += 1 # increase counter
  return ro # return positions


#@jit(nopython=True)
#@jit
def return_unique(rs1,rs2):
  """Return only those positions in rs1 that do not appear in rs2"""
  rout = []
  for ri in rs1:
    drs = [(ri-rj).dot(ri-rj) for rj in rs2] # distances
    if np.array(drs).min() > 0.001: rout.append(ri) # store this position
  return np.array(rout)




#def request(g,nat,ntries=20):
#  """Request a unit cell with as many atoms"""


def target_angle_volume(g,angle=None,n=5,volume=None,same_length=False):
    """Return a supercell, targetting a certain new angle between vectors"""
    if g.dimensionality!=2: raise # only for 2d
    a1 = g.a1
    a2 = g.a2
    def getm(): # get the matrix
      out = [] # empty list
      vs = [] # volumes
      for i in range(-n,n+1):
        for j in range(-n,n+1):
          for k in range(-n,n+1):
            for l in range(-n,n+1):
                store = False
                a1n = i*a1 + j*a2
                a2n = k*a1 + l*a2
                v = lg.norm(np.cross(a1n,a2n)) # new volume
                v /= lg.norm(np.cross(a1,a2)) # new volume
                if abs(v)<1e-6: continue # zero volume
                u1 = a1n/np.sqrt(a1n.dot(a1n)) # normalize
                u2 = a2n/np.sqrt(a2n.dot(a2n)) # normalize
                if angle is not None: # check if it has the desired angle
                    diff = u1.dot(u2)-np.cos(angle*np.pi) # difference
                    if abs(diff)>1e-6: continue # next try 
                if same_length: # check if they must have the same length
                    diff = a1n.dot(a1n) - a2n.dot(a2n)
                    if abs(diff)>1e-6: continue # next try 
                if volume is not None: # target such volume
                    if abs(v-volume)>1e-6: continue
                    else: return [[i,j,0],[k,l,0],[0,0,1]]
                out.append([[i,j,0],[k,l,0],[0,0,1]]) # orthogonal, return
                vs.append(v) # volume
      if len(out)==0: return None # nothng found
      vs = np.array(vs) # as array
      return [o for (v,o) in sorted(zip(vs,out))][0]
    out = getm() # get rotation matrix
    if out is None: raise
    g = g.supercell(out)
    g = sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.])) # set in the x direction
    return g


target_angle = target_angle_volume


def infer_supercell(g,g0):
    """Given two geometries, guess which supercell is associated"""
    # this only works for orthgonal supercells
    def norm(v): return np.sqrt(v.dot(v))
    if g.dimensionality==1:
        nx = int(norm(g.a1)/norm(g0.a1)) # out
        ny = 1
    elif g.dimensionality==2: # assume is orthogonal
        nx = int(norm(g.a1)/norm(g0.a1)) # out
        ny = int(norm(g.a2)/norm(g0.a2)) # out
    else: raise
    # probably a check should be added here
    return (nx,ny,1)
      








# compute certain order parameters
import numpy as np
from .. import multihopping

def st(h,nk=10,p=1):
    """Compute the singlet order parameter"""
    from . import extract
    if p==1: h = extract.get_singlet_hamiltonian(h)
    elif p==-1: h = extract.get_triplet_hamiltonian(h)
    else: raise
    hk = h.get_hk_gen() # get Bloch Hamiltonian generator
    ks = h.geometry.get_kmesh(nk=nk) # get k-mesh
    def f(k):
        m = hk(k)
        m = m@m
        return np.trace(m).real # return the trace
    return np.mean([f(k) for k in ks])

def singlet(h,**kwargs): return st(h,p=1,**kwargs)

def triplet(h,**kwargs): return st(h,p=-1,**kwargs)




import numpy as np
from .. import sculpt
from ..dos import calculate_dos


# specialized routines to compute the LDOS in continuum space

def ldosr_generator(h,rs=0.2,es=np.linspace(-1.,1.,100),
        delta=3e-2,nn=20,sector="electron",**kwargs):
    """Return a function that computes the LDOS as a function of the
    position"""
    (evals,vs) = h.get_eigenvectors(**kwargs,kpoints=False)
    ds = np.array([(np.conjugate(v)*v).real for v in vs]) # densities
    del vs # remove vectors
    g = h.geometry
    def fun(r):
        """Function to compute the DOS at position r"""
        r = np.array(r)
        inds = sculpt.get_closest(g,n=nn,r0=r) # indexes
        drs = [g.r[i]-r for i in inds] # distances
        dr = np.array([np.sqrt(ir.dot(ir)) for ir in drs]) # distances
        ws = np.exp(-dr/rs) # relative weight
        ws = ws/np.sum(ws) # normalize
        yout = 0. # initialize
        for i in range(len(inds)):
            ii = inds[i] # get this index
            if h.check_mode("spinless"):
              yi = calculate_dos(evals,es,delta,w=ds[:,ii])
              yout = yi*ws[i] # multiply by the weight
            elif h.check_mode("spinful"):
              yout = yout + calculate_dos(evals,es,delta,w=ds[:,2*ii])*ws[i]
              yout = yout+ calculate_dos(evals,es,delta,w=ds[:,2*ii+1])*ws[i]
            elif h.check_mode("spinful_nambu"):
              yout = yout + calculate_dos(evals,es,delta,w=ds[:,4*ii])*ws[i]
              yout = yout+ calculate_dos(evals,es,delta,w=ds[:,4*ii+1])*ws[i]
              if sector=="all":
                yout = yout + calculate_dos(evals,es,delta,w=ds[:,4*ii+2])*ws[i]
                yout = yout+ calculate_dos(evals,es,delta,w=ds[:,4*ii+3])*ws[i]
            else: raise
        return (es,yout)
    return fun










from ..geometry import Geometry
import numpy as np


def generate_keep_species(species):
    """Function to choose whether to keep a certain species"""
    if species is None: return lambda x: True # all
    def fkeep(s): return s in species
    return fkeep

def read_xyz(input_file="positions.xyz",species=None):
    """Read an xyz geometry"""
    fkeep = generate_keep_species(species) # function to decide species
    ls = open(input_file).readlines() # read the geometry
    del ls[0] # delete this line
    del ls[0] # delete this line
    rs = [] # empty list
    for i in range(len(ls)): # loop over lines
        l = ls[i].split() # split the line
        if len(l)<4: continue
        s = l[0]
        if fkeep(s): # check if this site should be kept
          x,y,z = float(l[1]),float(l[2]),float(l[3])
          rs.append([x,y,z])
    rs = np.array(rs) # convert to array
    from ..neighbor import neighbor_distances
    g = Geometry() # create geometry
    g.dimensionality = 0
    g.r = rs # positions
    g.r2xyz()
    g.normalize_nn_distance() # set the minimum distance to 1
    return g






import numpy as np
from numba import jit
from ..spectrum import eigenvalues_kmesh
from .. import parallel

# routines to compute the dielectric response function solely from eigenvalues
# note that this will ignore form factors coming from
# wavefunctions!


def epsilonk(h,energies=np.linspace(-4.0,4.0,100),nk=40,delta=2e-3,
        qs=None):
    """Compute the dielectric response function in reciprocal space"""
    es = eigenvalues_kmesh(h,nk=nk) # get eigenvalues in a mesh
    fp = lambda omega: epsilonk_fs_2d(es,omega,delta,qs=qs,nk=nk)
    out = parallel.pcall(fp,energies) # call in parallel
    return np.array(out)



def epsilonk_fs_2d(es,omega,delta,qs=None,nk=40):
    """Compute the response function in a 2d grid"""
    print("Doing",omega)
    out = epsilonk_2d_jit(es,omega,delta,np.zeros((nk,nk)))
    fs = fermisurface_2d_jit(es,omega,delta,np.zeros((nk,nk)))
    if qs is not None: 
        from ..interpolation import periodic_grid2mesh
        out = periodic_grid2mesh(out,qs)
        fs = periodic_grid2mesh(fs,qs)
    return out,fs



@jit(nopython=True)
def epsilonk_2d_jit(es,omega,delta,out):
    """Compute the response function in a 2d grid"""
    nx = es.shape[0] # nx
    ny = es.shape[1] # ny
    ne = es.shape[2] # number of energies
    for i in range(nx): # loop over kx
      for j in range(ny): # loop over ky
        for ii in range(nx): # loop over kx
          for jj in range(ny): # loop over ky
              o = 0.0 # initialize
              ei = es[(i+ii)%nx,(j+jj)%nx,:] # list of energies
              ej = es[i,j,:] # list of energies
              for ie in range(ne): # loop over energies
                for je in range(ne): # loop over energies
                    if ei[ie]<omega and ej[je]>omega:
                      o = delta/((ei[ie]-ej[je])**2+delta*delta) + o
              out[ii,jj] += o # add contribution
    return out # return output



@jit(nopython=True)
def fermisurface_2d_jit(es,omega,delta,out):
    """Compute the response function in a 2d grid"""
    nx = es.shape[0] # nx
    ny = es.shape[1] # ny
    ne = es.shape[2] # number of energies
    for i in range(nx): # loop over kx
      for j in range(ny): # loop over ky
         o = 0.0 # initialize
         ei = es[i,j,:] # list of energies
         for ie in range(ne): # loop over energies
            o = delta/((ei[ie]-omega)**2+delta*delta) + o
         out[i,j] = o # add contribution
    return out # return output





import numpy as np
from scipy.sparse import coo_matrix,csc_matrix,issparse


def add_phase(m1,r1,r2,phasefun,has_spin=False):
  if m1.shape[0] != len(r1): raise
  m = coo_matrix(m1) # convert to sparse matrix
  row,col = m.row,m.col
  data = m.data +0j
  for k in range(len(m.data)): # loop over non vanishing elements
    i = m.row[k]
    j = m.col[k]
    if has_spin: i,j = i//2,j//2 # if spinful
    # peierls phase
    p = phasefun(r1[i],r2[j]) # function yielding the phase
    data[k] *= p # add phase
  out = csc_matrix((data,(row,col)),shape=m1.shape) # convert to csc
  if not issparse(m1): out = out.todense() # dense matrix
  return out






def add_peierls(h,mag_field=0.0,new=False):
  """ Adds Peierls phase to the Hamiltonian"""
  if h.has_eh: raise
  x = h.geometry.x    # x coordinate 
  y = h.geometry.y    # x coordinate 
  a1 = h.geometry.a1    # distance to neighboring cell
  celldis = np.sqrt(a1.dot(a1))
  from numpy import array
  norb = h.intra.shape[0]  # number of orbitals
  if h.is_multicell: raise
  if new:
    print("New method to calculate peierls")
    g = h.geometry # geometry
    has_spin = h.has_spin # has spin degree of freedom
    def phasefun(ri,rj): 
      if callable(mag_field): return mag_field(ri[0],ri[1],rj[0],rj[1])
      else: raise
    h.intra = add_phase(h.intra,g.r,g.r,phasefun,has_spin) 
    if h.dimensionality==2:
      h.tx = add_phase(h.tx,g.r,g.replicas([1.,0.,0.]),phasefun,has_spin) 
      h.ty = add_phase(h.ty,g.r,g.replicas([0.,1.,0.]),phasefun,has_spin) 
      h.txy = add_phase(h.txy,g.r,g.replicas([1.,1.,0.]),phasefun,has_spin) 
      h.txmy = add_phase(h.txmy,g.r,g.replicas([1.,-1.,0.]),phasefun,has_spin) 
      return
  else: # old method


    if h.is_sparse: # sparse hamiltonian
      if True: # zero dimensional
        m = coo_matrix(h.intra) # convert to sparse matrix
        row,col = m.row,m.col
        data = m.data +0j
        for k in range(len(m.data)): # loop over non vanishing elements
          i = m.row[k]
          j = m.col[k]
          if h.has_spin: i,j = i/2,j/2 # raise if spinful
          p = peierls(x[i],y[i],x[j],y[j],mag_field) # peierls phase
          data[k] *= p # add phase
        h.intra = csc_matrix((data,(row,col)),shape=(norb,norb)) # convert to csc
      if h.dimensionality==1: # one dimensional
        # check that celldis is right
        if np.abs(celldis - h.geometry.a1[0])>0.001: raise
        def phaseize(inter,numn=1):
          m = coo_matrix(inter) # convert to sparse matrix
          row,col = m.row,m.col
          data = m.data +0j
          for k in range(len(m.data)): # loop over non vanishing elements
            i = m.row[k]
            j = m.col[k]
            if h.has_spin: i,j = i//2,j//2 # raise if spinful
            # peierls phase
            p = peierls(x[i],y[i],x[j]+numn*celldis,y[j],mag_field) 
            data[k] *= p # add phase
          return csc_matrix((data,(row,col)),shape=(norb,norb)) # convert to csc
        # for normal hamiltonians
        if not h.is_multicell: h.inter = phaseize(h.inter,numn=1) 
        # for multicell hamiltonians
        if h.is_multicell: 
          hopping = [] # empty list
          for i in range(len(h.hopping)):
            h.hopping[i].m = phaseize(h.hopping[i].m,numn=h.hopping[i].dir[0]) 
    else: # not sparse
      def gaugeize(m,d=0.0):
        """Add gauge phase to a matrix"""
        for i in range(len(x)):
          for j in range(len(x)):
            p = peierls(x[i],y[i],x[j]+d,y[j],mag_field) # peierls phase
            if h.has_spin:
              m[2*i,2*j] *= p
              m[2*i,2*j+1] *= p
              m[2*i+1,2*j] *= p
              m[2*i+1,2*j+1] *= p
            else:
              m[i,j] *= p
      gaugeize(h.intra,d=0.0)  # gaugeize intraterm
      if h.dimensionality==0: pass # if zero dimensional
      elif h.dimensionality==1: # if one dimensional
        if h.is_multicell: raise
        gaugeize(h.inter,d=celldis) # gaugeize interterm
      elif h.dimensionality==2: # if bigger dimensional
        print("WARNING, is your gauge periodic?")
        gaugeize(h.tx,d=h.geometry.a1[0]) # gaugeize interterm
      else:
        raise




def peierls(x1,y1,x2,y2,mag_field):
  """ Returns the complex phase with magnetic field """
  if is_number(mag_field): b = mag_field 
  elif callable(mag_field): 
      b = mag_field(np.array([x1,y1,0.0]),np.array([x2,y2,0.0]))
  else: raise
  phase = b*(x1-x2)*(y1+y2)/2.0
  return np.exp(1j*phase)



def is_number(s):
    try:
        float(s)
        return True
    except:
        return False




def add_inplane_bfield(h,**kwargs):   add_bfield(h,mode="inplane",**kwargs)
def add_offplane_bfield(h,**kwargs):   add_bfield(h,mode="offplane",**kwargs)

def add_peierls(h,mag_field,**kwargs):
    add_offplane_bfield(h,b=mag_field*2,**kwargs)





def add_bfield(h,b=0.0,phi=0.0,mode="inplane",gauge="Landau"):
    """Add an in-plane magnetic field"""
    if h.dimensionality>2: raise # not implemented
    if h.has_spin: gi = lambda i: i//2
    else: gi = lambda i: i
    if not h.is_multicell: 
        h.turn_multicell() # turn to multicell form
    cphi = np.cos(phi*np.pi)
    sphi = np.sin(phi*np.pi)
    g = h.geometry # get geometry
    r = g.r # get positions
    def get_phase(r,dr):
        if mode=="inplane": return 2*r[2]*(dr[0]*sphi - dr[1]*cphi)
        elif mode=="offplane": 
            if gauge=="Landau": return r[1]*dr[0]/2.
            elif gauge=="symmetric": return r[1]*dr[0]/4.- r[0]*dr[1]/4.
            else: raise
        else: raise
    def add_phase(m,r1,r2): # add the phase
        mo = coo_matrix(m) # convert to coo matrix
        data = mo.data +0.0j
        k = 0
        for (i,j,d) in zip(mo.row,mo.col,mo.data): # loop
            r = (r1[gi(i)] + r2[gi(j)])/2.
            dr = r1[gi(i)] - r2[gi(j)]
            p = get_phase(r,dr)
            data[k] *= np.exp(1j*b*p)
            k += 1
        out = csc_matrix((data,(mo.row,mo.col)),shape=mo.shape) 
        return out.todense() # return matrix
    # now add the phase to the Hamiltonian
    h.intra = add_phase(h.intra,r,r) # add the phase
    for i in range(len(h.hopping)):
        t = h.hopping[i]
        h.hopping[i].m = add_phase(t.m,r,g.replicas(t.dir))
    return h







import numpy as np
from scipy.interpolate import interp1d

def positive_interpolator(x,y,fill=None):
    """Given x and y, return a function that interpolates
    data, assuming a positive interpolation"""
    if fill is None: fill = (y[0],y[len(y)-1]) # filled value
    f = interp1d(x,y,kind="linear",bounds_error=False,
                    fill_value=fill) # interpolation
    return lambda x: f(x)*(1.+np.sign(f(x)))/2. # only positive




import numpy as np
from copy import deepcopy
from ..superconductivity import get_eh_sector
from ..superconductivity import build_nambu_matrix



# different constrains for the meanfield calculation

def remove_onsite_spinful(m):
    n = m.shape[0]//2 # number of orbitals
    for i in range(n):
        d = (m[2*i,2*i] + m[2*i+1,2*i+1])/2. # average charge
        m[2*i,2*i] -= d # set
        m[2*i+1,2*i+1] -= d # set
    return m



def remove_magnetism_spinful(m):
    n = m.shape[0]//2 # number of orbitals
    for i in range(n):
        d = (m[2*i,2*i] + m[2*i+1,2*i+1])/2. # average charge
        m[2*i,2*i] = d # set
        m[2*i+1,2*i+1] = d # set
        m[2*i,2*i+1] = 0. # set
        m[2*i+1,2*i] = 0. # set
    return m


def remove_offplane_magnetism_spinful(m):
    n = m.shape[0]//2 # number of orbitals
    for i in range(n):
        d = (m[2*i,2*i] + m[2*i+1,2*i+1])/2. # average charge
        m[2*i,2*i] = d # set
        m[2*i+1,2*i+1] = d # set
    return m



def remove_inplane_magnetism_spinful(m):
    n = m.shape[0]//2 # number of orbitals
    for i in range(n):
        m[2*i,2*i+1] = 0. # set
        m[2*i+1,2*i] = 0. # set
    return m




def remove_spinful_sector(h,removef):
    """Remove total charge renormalization"""
    has_eh = h.has_eh
    has_spin = h.has_spin
    def f(dd): # create function
        out = deepcopy(dd) # copy the dictionary
        m = out[(0,0,0)] # onsite matrix
        if has_eh and not has_spin: raise # not implemented
        elif not has_eh and has_spin: # spinful
            m = removef(m)
        elif has_eh and has_spin: # spinful
            m01 = get_eh_sector(m,i=0,j=1) # anomalous part
            m10 = get_eh_sector(m,i=1,j=0) # anomalous part
            m00 = get_eh_sector(m,i=0,j=0) # anomalous part
            m00 = removef(m00) # remove onsite 
            m = build_nambu_matrix(m00,c12=m01,c21=m10) # rebuild the matrix
        else: raise
        out[(0,0,0)] = m # set the new matrix
        return out # return dictionary
    return f # return function


def remove_charge(h):
    return remove_spinful_sector(h,remove_onsite_spinful)

def remove_magnetism(h):
    return remove_spinful_sector(h,remove_magnetism_spinful)

def remove_inplane_magnetism(h):
    return remove_spinful_sector(h,remove_inplane_magnetism_spinful)

def remove_offplane_magnetism(h):
    return remove_spinful_sector(h,remove_offplane_magnetism_spinful)




def enforce_constrains(mf,h,constrains=[]):
    """Given a list of constrains, return a function that enforces
    all of them in the mean field"""
    for c in constrains:
        if c=="no_charge":
            mf = remove_charge(h)(mf) # remove charge renormalization
        elif c=="no_magnetism":
            mf = remove_magnetism(h)(mf) # remove magnetism
        elif c=="no_inplane_magnetism":
            mf = remove_inplane_magnetism(h)(mf) # remove inplane magnetism
        elif c=="no_offplane_magnetism":
            mf = remove_offplane_magnetism(h)(mf) # remove inplane magnetism
        elif c=="no_normal_term":
            if h.has_eh:
                from ..sctk.extract import extract_anomalous_dict
                mf = extract_anomalous_dict(mf)
            else: raise
        elif c in ["no_anomalous_term","no_SC","no_superconductivity"]:
            if h.has_eh:
                from ..sctk.extract import extract_normal_dict
                mf = extract_normal_dict(mf)
            else: pass
#        print(np.round(mf[(0,0,0)],1))
    return mf




def obj2mf(mf):
    from ..algebra import ismatrix
    from ..hamiltonians import Hamiltonian
    if ismatrix(mf): return {(0,0,0):mf}
    elif type(mf)==Hamiltonian: return mf.get_dict()
    else: return mf # assume it is a valid dictionary




import numpy as np
from . import classicalspinf90
from . import neighbor
from scipy.sparse import csr_matrix,csc_matrix,coo_matrix

zero = np.matrix(np.zeros((3,3)))  # real matrix
iden = np.matrix(np.identity(3))  # real matrix
zzm = zero.copy() ; zzm[2,2] = 1.0 # matrix for ZZ interaction


class SpinModel(): # class for a spin Hamiltonian
  def __init__(self,g): # geometry
    self.geometry = g # store geometry
    self.nspin = len(g.r) # number of sites
    self.theta = np.zeros(len(g.r)) # theta angles
    self.phi = np.zeros(len(g.r)) # phi angles
    self.b = np.zeros((len(g.r),3)) # magnetic field
    self.j = np.array([zero]) # empty list
    self.pairs = np.array([[0,0]]) # empty list
  def add_heisenberg(self,fun=None):
    h = self.geometry.get_hamiltonian(has_spin=False,fun=fun)
    m = coo_matrix(h.get_hk_gen()([0.,0.,0.])) # get onsite matrix
    (pairs,js) = add_heisenberg(self.geometry.r)
    pairs = np.array([m.row,m.col]).transpose() # convert to array
    self.pairs = pairs
    self.j = [ j.real*np.identity(3) for j in m.data] # store
  def add_field(self,v):
    """Add magnetic field"""
    self.b += np.array([v for i in range(self.nspin)])
  def energy(self,use_fortran=True):
    """ Calculate the energy"""
    if use_fortran:
      eout = classicalspinf90.energy(self.theta,self.phi,self.b,self.j,
               self.pairs)
    return eout
  def minimize_energy(self,theta0=None,phi0=None,tries=10,calle=None):
    """Minimize the energy of the spin model"""
    thetas = [None for i in range(tries)]
    phis = [None for i in range(tries)]
    es = [None for i in range(tries)]
    for i in range(tries): # loop over tries
      theta,phi = minimize_energy(self,theta0=theta0,phi0=phi0,calle=calle)
      self.theta = theta.copy()
      self.phi = phi.copy()
      e = self.energy() # calculate energy
      thetas[i] = theta.copy()
      phis[i] = phi.copy()
      es[i] = e
    imin = es.index(min(es)) # minimum
    print("Minimum energy",es[imin])
    print("Minimum energy per spin",es[imin]/len(self.phi))
    self.theta = thetas[imin].copy()
    self.phi = phis[imin].copy()
    self.update_magnetization()
    return self.theta,self.phi
  def write(self,label=""):
    """Write in file"""
    self.geometry.write() # write the geometry
    write_magnetization(self) # write magnetization
  def add_tensor(self,fun):
    """Add a tensor interaction"""
    pairs,js = add_tensor(self,fun)
    self.pairs = np.concatenate([self.pairs,pairs])
    self.j = np.concatenate([self.j,js])
  def add_tensor_2d(self,fun,ncells=1,vspiral=[0.,0.]):
    pairs,js = add_tensor_2d(self,fun,ncells=ncells,vspiral=vspiral)
    self.pairs = np.concatenate([self.pairs,pairs])
    self.j = np.concatenate([self.j,js])
  def get_jacobian(self):
    """Return a function that calculated the Jacobian"""
#    return None
    return get_jacobian(self.b,self.j,self.pairs)
  def load_magnetism(self,name="MAGNETIZATION.OUT"):
    """Read the magnetization from a file"""
    m = np.genfromtxt(name).transpose()
    r = np.sqrt(m[1]**2+m[2]**2) # in-plane radius
    self.theta = np.arctan2(r,m[3]) # theta angle
    self.phi = np.arctan2(m[2],m[1]) # theta angle
  def regroup(self):
    """Regroups the terms in the Hamiltonian"""
    print(len(self.pairs))
    self.pairs,self.j = regroup(self.pairs,self.j) 
    print(len(self.pairs))
  def update_magnetization(self):
    """Update the magnetization"""
    mx = np.sin(self.theta)*np.cos(self.phi)
    my = np.sin(self.theta)*np.sin(self.phi)
    mz = np.cos(self.theta)
    self.magnetization = np.array([mx,my,mz]).transpose()
  def setup_matrix(self):
    """Creates the matrix that allows to calculate the energy""" 




def add_heisenberg(r):
  """Return pairs and js for a Heisenberg interaction"""
  pairs = neighbor.find_first_neighbor(r,r)
  js = np.array([iden for p in pairs]) # array
  return (pairs,js) # return pairs 


def energy(thetas,phis,bs,js,indsjs):
  """Calculate the energy"""
  eout = classicalspinf90.energy(thetas,phis,bs,js,indsjs)
  return eout



def get_jacobian(bs,js,indsjs):
  """Return the Jacobian"""
  def jacobian(thetaphi):
    thetas = thetaphi[0:len(thetaphi)//2]
    phis = thetaphi[len(thetaphi)//2:len(thetaphi)]
    jac = classicalspinf90.jacobian(thetas,phis,bs,js,indsjs)
    return jac
  return jacobian


def minimize_energy(sm,theta0=None,phi0=None,tol=1e-5,calle=None):
  """Return the minimum energy and gound state theta/phi"""
  if theta0 is None: theta0 = np.random.random(sm.nspin)*np.pi
  if phi0 is None: phi0 = np.random.random(sm.nspin)*np.pi*2
  from scipy.optimize import minimize
  thetaphi0 = np.concatenate([theta0,phi0]) # initial guess
  if calle is None: # no added function to the total energy
    def fun(thetaphi): # function to optimize
      sm.theta = thetaphi[0:sm.nspin]
      sm.phi = thetaphi[sm.nspin:sm.nspin*2]
      return sm.energy() # return energy
    result = minimize(fun,thetaphi0,tol=tol,jac=sm.get_jacobian())
  else: # a function is provided that corrects the energy
    def fun2(thetaphi): # function yielding the energy
      sm.theta = thetaphi[0:sm.nspin]
      sm.phi = thetaphi[sm.nspin:sm.nspin*2]
      sm.update_magnetization() # update the magnetization
      return sm.energy() + calle(sm) # return total energy
    result = minimize(fun2,thetaphi0,tol=tol,method="Powell") # minimize this function
  theta = result.x[0:sm.nspin]
  phi = result.x[sm.nspin:sm.nspin*2]
  return theta,phi

def write_magnetization(self):
  theta = self.theta
  phi = self.phi
  mx = np.sin(theta)*np.cos(phi)
  my = np.sin(theta)*np.sin(phi)
  mz = np.cos(theta)
  fo = open("MAGNETISM.OUT","w")
  for i in range(len(mx)):
    fo.write(str(self.geometry.x[i])+"   ")
    fo.write(str(self.geometry.y[i])+"   ")
    fo.write(str(self.geometry.z[i])+"   ")
    fo.write(str(mx[i])+"   ")
    fo.write(str(my[i])+"   ")
    fo.write(str(mz[i])+"\n")
  fo.close()



def add_tensor(sm,fun):
  """Add a tensor interaction"""
  pairs = []
  js = []
  for i1 in range(sm.nspin): # loop over positions
    r1 = sm.geometry.r[i1]
    for i2 in range(sm.nspin): # loop over positions
      r2 = sm.geometry.r[i2]
      m = fun(r1,r2) # call the function 
      if np.max(np.abs(m))>0.00001: # if non zero
        pairs.append((i1,i2)) # store 
        js.append(m) # store matrix
  return np.array(pairs),np.array(js)


def rotatez(angle):
  """In-plane rotation"""
  m = np.matrix(np.zeros((3,3),dtype=np.complex)) # zero matrix
  cp = np.cos(angle*np.pi)
  sp = np.sin(angle*np.pi)
  m[0,0] = cp
  m[1,1] = cp
  m[0,1] = sp
  m[1,0] = -sp
  m[2,2] = 1.0
  return m # return matrix





def add_tensor_2d(sm,fun,ncells=4,vspiral=[0.,0.]):
  """Add a tensor interaction, assuming 2d system"""
  pairs = []
  js = []
  for ia in range(-ncells,ncells+1): # loop over cells
    for ja in range(-ncells,ncells+1): # loop over cells
      for i1 in range(sm.nspin): # loop over positions
        r1 = sm.geometry.r[i1]
        for i2 in range(sm.nspin): # loop over positions
          r2 = sm.geometry.r[i2] + ia*sm.geometry.a1 + ja*sm.geometry.a2
          m = fun(r1,r2) # call the function 
          if np.max(np.abs(m))>0.00001: # if non zero
            angle = vspiral[0]*ia + vspiral[1]*ja
            R = rotatez(angle)
            pairs.append((i1,i2)) # store 
            js.append(m*R) # store matrix
  return np.array(pairs),np.array(js)








def generating_functions(name="Heisenberg",J=1.0,v=np.array([0.,0.,1.]),
       fdiff = lambda x,y: x-y,fc=None,fr=None):
  """Return fucntion that return matrices for spin itneractions"""
  if fc is None: # if no function provided
    def fc(d):
      if 0.9<d<1.1: return 1.0
      else: return 0.0
  if fr is None: # if no function provided
    def fr(r1,r2): # vector dependent rotation
      return iden # identity matrix
  if name=="Heisenberg":
    def fun(r1,r2):
      dr = fdiff(r2,r1)
      dr = np.sqrt(dr.dot(dr))
      return iden*J*fc(dr)*fr(r1,r2) # distance dependent coupling
    return fun
  elif name=="Linear":
    def fun(r1,r2):
      r12 = fdiff(r2,r1) # vector between them
      dr = np.sqrt(r12.dot(r12)) # distance
      if dr<0.1: return np.zeros((3,3))
      ur = r12/dr # unit vector
      m = np.matrix([[ur[i]*ur[j] for i in range(3)] for j in range(3)])
      return m*J*fr(r1,r2)/dr**3 # distance dependent coupling
    return fun
  elif name=="RKKYTI":
    """RKKY interaction in the surface of a TI, as derived in
    PRB 81 233405 (2010) """
    def fun(r1,r2):
      r12 = fdiff(r2,r1) # vector between them
      dr = np.sqrt(r12.dot(r12)) # distance
      if dr<0.1: return np.zeros((3,3)) # return 0
      ur = r12/dr # unit vector
      m = np.matrix([[ur[i]*ur[j] for i in range(3)] for j in range(3)])
      m = m*3/2 - np.identity(3) 
      return m*J*fr(r1,r2)/dr**3 # distance dependent coupling
    return fun
  elif name=="ZZ":
    def fun(r1,r2):
      dr = fdiff(r2,r1)
      dr = np.sqrt(dr.dot(dr))
      return zzm*J*fc(dr)*fr(r1,r2) # return identity
    return fun
  elif name=="XYZ":
    def fun(r1,r2):
      dr = fdiff(r2,r1)
      dr2 = np.sqrt(dr.dot(dr))
      if np.abs(fc(dr2))<0.00000001: return zero
      if callable(v): return J*fc(dr2)*np.diag(v(dr)) # return matrix
      else: return J*fc(dr)*np.diag(v)*fr(r1,r2) # return matrix
    return fun
  elif name=="DM":
    eps = get_lc()
    def fun(r1,r2):
      dr1 = fdiff(r2,r1)
      dr = np.sqrt(dr1.dot(dr1))
      if 0.9<dr<1.1: 
        m = np.zeros((3,3)) # intialize the matrix
        if callable(v): rm = np.cross(dr1,v(dr1)) # intermediate ion
        else: rm = np.cross(dr1,v) # intermediate ion
        rm = rm/np.sqrt(rm.dot(rm)) # unitary vector
        for i in range(3):
          for j in range(3): 
            for k in range(3): 
              m[i,j] += eps[i,j,k]*rm[k]
        return m*J*fr(r1,r2) # return identity
      else: return zero
    return fun



def generating_profiles(r,name="skyrmion",n=1.,cut=1.0):
  """Return thetas and phis"""
  rho = np.sqrt(r[:,0]**2 + r[:,1]**2) # in-plnae r
  if name=="skyrmion":
    theta = cut*rho/np.max(rho)*np.pi
    phi = n*np.arctan2(r[:,1],r[:,0])
    return theta,phi
  elif name=="spiral":
    phi = (r[:,0]/np.max(r[:,0])+1)*n*np.pi*2
    return phi*0.+np.pi/2,phi
  else: raise


def get_lc():
  """Return a Levi Civita"""
  e = np.zeros((3,3,3)) # start as zero
  e[0,1,2] = 1.0 
  e[0,2,1] = -1.0 
  e[1,0,2] = -1.0 
  e[1,2,0] = 1.0 
  e[2,0,1] = 1.0 
  e[2,1,0] = -1.0 
  return e





def regroup(ps,js):
  """Regroups the terms in the Hamiltonian"""
  outp = [] # output pairs
  outj = [] # output js
  dictj = dict() # create dictionary
  for (p,j) in zip(ps,js): # loop over inputs
    ip = (p[0],p[1])
    jp = (p[1],p[0])
    if ip in outp: # if the pair is present
      dictj[ip] += j # add contribution
    elif jp in outp: # if the alternative pair is present
      dictj[jp] += j.transpose() # add contribution
    else: # not present
      outp.append(ip) # store
      dictj[ip] = j # store this j
  outj = [dictj[p] for p in outp] # get all
  return np.array(outp),np.array(outj)





from ..greentk.selfenergy import bloch_selfenergy


def get_gf(H,energy=0.0,delta=1e-5,
           gtype="bulk",
           mode="adaptive",
           **kwargs):
    """Generate the specific Green's function"""
    if H.dimensionality>0:
        gf = bloch_selfenergy(H,energy=energy,
                                         mode=mode,
                                         gtype=gtype,
                                         delta=delta)[0]
        return gf
    elif H.dimensionality==0: # zero dimensional
        return H.get_gk_gen(delta=delta)(e=energy) 
    else: raise




import numpy as np
from .. import algebra
from .. import densitymatrix

def real_space_chern(h,operator=None):
    """Compute the real space Chern number"""
    if h.dimensionality!=0: raise
    X = h.get_operator("xposition").get_matrix()
    Y = h.get_operator("yposition").get_matrix()
#    X = np.diag(h.geometry.x)
#    Y = np.diag(h.geometry.y)
    m = h.get_hk_gen()([0.,0.,0.]) # get the matrix
    P = densitymatrix.occupied_projector(m).T # projector on occupied states
    Q = np.identity(P.shape[0]) - P # projector on unoccupied states
#    P = np.conjugate(P.T)
    r = h.geometry.r
    dr = np.array([ri.dot(ri) for ri in r])
    rcut = np.max(dr)/3.
    scale = np.pi*rcut/len(dr[dr<rcut]) # scaling for the Chern number
#    Ph = np.conjugate(P.T) # dagger
    A,B = P@X@P,P@Y@P # define the two operators
    if operator is not None: # in case there is a certain projector
        op = operator.get_matrix()
        C1 = A@op@B - B@A@op # compute the commutator
        C2 = A@B@op - B@op@A # compute the commutator
        C = (C1+C2)/2. # average
    elif operator is None: C = A@B - B@A # compute the commutator
    else: raise
    C = np.pi*2*np.diagonal(C).imag # diagonal part
    C = C/scale # normalize
#    print(P - P@P)
    C = h.full2profile(C) # resum if necessary
    h.geometry.write_profile(C,name="REAL_SPACE_CHERN.OUT")









from __future__ import print_function
import numpy as np
import pylab as py
from copy import deepcopy as dc
import scipy.linalg as lg
from scipy.sparse import bmat,coo_matrix,csc_matrix
from . import green
from .algebra import dagger

dag = dagger



class Heterostructure():
  """Class for a HTstructure"""
  def copy(self):
    """Copy the HTstructure"""
    from copy import deepcopy
    return deepcopy(self)
  def __init__(self,h=None):  # initialization using a hamiltonian
    self.file_right_green = "green_right.dat"  # in/out file for right green
    self.file_left_green = "green_left.dat"   # in/out file for left green
    self.file_heff = "heff.dat"   # out for the effective hamiltonian
    self.scale_rc = 1.0 # scale coupling to the right lead
    self.scale_lc = 1.0 # scale coupling to the left lead
    self.is_sparse = False
    self.dimensionality = 1 # default is one dimensional
    self.transparency = 1.0 # reference transparency (for kappa)
    self.delta = 0.0001
    self.extra_delta_central = 0. # additional delta in the central region
    self.extra_delta_right = 0. # additional delta in the right region
    self.extra_delta_left = 0. # additional delta in the left region
    self.interpolated_selfenergy = False
    self.block_diagonal = False
    if h is not None:
      self.heff = None  # effective hamiltonian
      self.right_intra = h.intra  # intraterm in the right lead
      self.right_inter = h.inter  # interterm in the right lead (to the right)
      self.right_green = None  # right green function
      self.left_intra = h.intra  # intraterm in the left lead
      self.left_inter = dagger(h.inter)  # interterm in the left lead (to the left)
      self.left_green = None  # left green function
      self.central_intra = h.intra  # intraterm in the center
      self.right_coupling = h.inter # coupling from the center to the right lead
      self.left_coupling = dagger(h.inter) # coupling from the center to the left lead
      # geometry of the central part
      gc = dc(h.geometry)
      self.central_geometry = gc # geometry of the central part
      # additional degrees of freedom
      self.has_spin = h.has_spin   # spin degree of freedom
      self.has_eh = h.has_eh   # electron hole pairs
#  def plot_central_dos(self,energies=[0.0],num_rep=100,
#                      mixing=0.7,eps=0.0001,green_guess=None,max_error=0.0001):
#    """ Plots the density of states by using a 
#    green function approach"""
#    return plot_central_dos(self,energies=energies,num_rep=num_rep,
#                      mixing=mixing,eps=eps,green_guess=green_guess,
#                      max_error=max_error)
  def surface_dos(self,**kwargs):
      from .transporttk import sdos
      return sdos.surface_dos(self,**kwargs)
  def get_coupled_central_dos(self,**kwargs):
      return device_dos(self,mode="central",**kwargs)
  def get_coupled_left_dos(self,**kwargs):
      return device_dos(self,mode="left",**kwargs)
  def get_coupled_right_dos(self,**kwargs):
      return device_dos(self,mode="right",**kwargs)
  def landauer(self,energy=[0.],delta=0.0001,do_leads=True,left_channel=None,
                right_channel=None):
    """ Return the Landauer transmission"""
    if self.has_eh: raise # invalid if there is electorn-hole
    return landauer(self,energy=energy,delta=delta,do_leads=do_leads,
                    left_channel=left_channel,right_channel=right_channel)
  def write_green(self):
    """Writes the green functions in a file"""
    from .green import write_matrix
    write_matrix(self.file_right_green,self.right_green)
    write_matrix(self.file_left_green,self.left_green)
  def read_green(self):
    """Reads the green functions from a file"""
    from .green import read_matrix
    self.right_green = read_matrix(self.file_right_green)
    self.left_green = read_matrix(self.file_left_green)
  def write_heff(self):
    """ Writes effective hamiltonian in a file"""
    from .green import write_sparse
    write_sparse(self.file_heff,self.heff)
  def eigenvalues(self,numeig = 10,effective=False,full=False):
    """ Calculates eigenvalues of the central part """
    return eigenvalues(self,numeig=numeig,effective=effective,
                        full=full)
  def replace_center(self,ht_replacement):
      """ Replaces the central part by the second argument"""
      self.central_intra = ht_replacement.central_intra  # make the change
  def calculate_surface_green(self,energy=0.0,delta=0.0001,error=0.0000001):
      """Calculate the surface Green function"""
      delta = self.delta
      # Right
      intra = self.right_intra
      inter = self.right_inter
      gbulk,g = green.green_renormalization(intra,inter,error=error,
                                             energy=energy,delta=delta)
      self.right_green = g # save green function
      # Left
      intra = self.left_intra
      inter = self.left_inter
      gbulk,g = green.green_renormalization(intra,inter,error=error,
                                             energy=energy,delta=delta)
      self.left_green = g # save green function
      self.energy_green = energy # energy of the Green function
  def copy_surface_green(self,ht):
      """Copy the surface Green fucntions"""
      self.energy_green = ht.energy_green
      self.left_green = ht.left_green
      self.right_green = ht.right_green
  def get_selfenergy(self,energy,**kwargs):
     """Return selfenergy of iesim lead"""
     from .transporttk.selfenergy import get_selfenergy
     return get_selfenergy(self,energy,**kwargs)
  def get_reflection_normal_lead(self,s):
     from .transporttk.builder import get_reflection_normal_lead
     return get_reflection_normal_lead(self,s)
  def get_central_gmatrix(self,**kwargs):
     """Return the inverse central Green's function"""
     from .transporttk.smatrix import get_central_gmatrix
     return get_central_gmatrix(self,**kwargs) 
  def set_coupling(self,c): 
     """Coupling for kappa functionality"""
     self.scale_rc = c
  def setup_selfenergy_interpolation(self,es=np.linspace(-4.0,4.0,100),
           delta=0.0001,pristine=False):
    """Create the functions that interpolate the selfenergy"""
    from .interpolation import intermatrix
    self.interpolated_selfenergy = False # set as False
    fsl = lambda e: self.get_selfenergy(e,delta=delta,lead=0,pristine=pristine)
    fsr = lambda e: self.get_selfenergy(e,delta=delta,lead=1,pristine=pristine)
    fun_sr = intermatrix(fsr,xs=es) # get the function
    fun_sl = intermatrix(fsl,xs=es) # get the function
    self.selfgen = [fun_sl,fun_sr] # store functions
    self.interpolated_selfenergy = True # set as true
  def didv(self,**kwargs):
      from .transporttk.didv import generic_didv
      return generic_didv(self,**kwargs)
  def block2full(self,sparse=False):
      """Put in full form"""
      return block2full(self,sparse=sparse)
  def get_kappa(self,**kwargs):
      from .transporttk.kappa import get_kappa_ratio
      return get_kappa_ratio(self,**kwargs)
  




def create_leads_and_central(h_right,h_left,h_central,num_central=1,
      interpolation="None",block_diagonal=True):
  """ Creates an HTjunction by giving the hamiltonian
     of the leads and the center, with a certain number of cells
     in the center  """
  # check the hamiltonians
  if num_central==1: block_diagonal = False
#  h_right.check()
#  h_left.check()
#  h_central.check()
  ht = HTstructure(h_central) # create HTstructure
  ht.has_eh = h_right.has_eh # if it has electron-hole
  ht.get_eh_sector = h_right.get_eh_sector # if it has electron-hole
  # assign matrices of the leads
  ht.right_intra = h_right.intra.copy()  
  ht.right_inter = h_right.inter.copy()  
  ht.left_intra = h_left.intra.copy()  
  ht.left_inter = dagger(h_left.inter).copy() 
  # create matrix of the central part and couplings to the leads
  from scipy.sparse import csc_matrix,bmat
  z = csc_matrix(h_central.intra*0.0j) # zero matrix
  hc = [[None for i in range(num_central)] for j in range(num_central)]
  tcr = [[None] for i in range(num_central)]
  tcl = [[None] for i in range(num_central)]  
  ##############################################################
  # create the centrla hamiltonian according to different schemes
  ##############################################################
  if not block_diagonal: 
    ec = csc_matrix(h_central.intra) 
    er = csc_matrix(h_right.intra) 
    el = csc_matrix(h_left.intra) 
    tr = csc_matrix(h_right.inter) 
    tl = csc_matrix(h_right.inter) 
    tc = csc_matrix(h_central.inter) 
  if block_diagonal: 
    ec = h_central.intra.copy()  
    er = h_right.intra.copy()  
    el = h_left.intra.copy()  
    tr = h_right.inter.copy()  
    tl = h_right.inter.copy() 
    tc = h_central.inter.copy() 
  # central part is pure central input hamilotnian
  if interpolation=="None": # without central interpolation
    for i in range(num_central):  # intra term of the central blocks
      hc[i][i] = ec
    for i in range(num_central-1): # interterm of the central blocks
      hc[i][i+1] = tc
      hc[i+1][i] = tc.H

  # central part is a step of the right and left hamiltonians
  elif interpolation=="step": # with step interpolation
    # intraterm
    for i in range(num_central):  # intra term of the central blocks
      if i<num_central/2:
          hc[i][i] = el
      elif i>num_central/2:
          hc[i][i] = er
      elif i==num_central/2:
          hc[i][i] = ec
      else:
        raise
    # interterm
    for i in range(num_central-1): # interterm of the central blocks
      if i<num_central/2:
        hc[i][i+1] = tl
        hc[i+1][i] = tl.H
      elif i>num_central/2:
        hc[i][i+1] = tr
        hc[i+1][i] = tr.H
      elif i==num_central/2:
        hc[i][i+1] = tc
        hc[i+1][i] = tc.H
      else:
        raise


  # central part is a linear interpolation of right and left
  elif interpolation=="linear": 
    for i in range(num_central):  # intra term of the central blocks
      r = float(i+1)/float(num_central+1)   # 
      hc[i][i] = er*r + el*(1.-r)  # mean interpolation
    for i in range(num_central-1): # interterm of the central blocks
      r = float(i+1)/float(num_central+1)   # 
      hc[i][i+1] = tr*r + tl*(1.-r)
      hc[i+1][i] = (tr*r + tl*(1.-r)).H


  else:
    raise

  for i in range(num_central):  # intra term of the central blocks
    tcr[i][0] = csc_matrix(z) 
    tcl[i][0] = csc_matrix(z) 

  tcr[-1][0] = csc_matrix(h_right.inter) # hopping to the right lead
  tcl[0][0] = csc_matrix(dagger(h_left.inter)) # hopping to the left lead
  # create dense matrices
  if not block_diagonal:
    hc = bmat(hc).todense()
    tcr = bmat(tcr).todense() 
    tcl = bmat(tcl).todense()
    ht.block_diagonal = False
  # do not create dense matrices
  if block_diagonal:
    hc = hc   # this is a list !!!!!
    tcr = h_right.inter.copy()    # this is a matrix
    tcl = h_left.inter.H.copy()   # this is a matrix
    ht.block_diagonal = True
  # assign to the HTstructure
  ht.right_coupling = tcr
  ht.left_coupling = tcl
  ht.central_intra = hc
  # and modify the geometry
  ht.central_geometry.supercell(num_central) 
  return ht

HTstructure = Heterostructure

def device_dos(HT,energy=0.0,mode="central",operator=None):
   """ Calculates the density of states 
       of a HTstructure by a  
       green function approach, input is a HTstructure class
   """
   from .transporttk import fullgreen
   g = fullgreen.get_full_green(HT,energy,mode=mode)
   if operator is not None:
       g = HT.Hr.get_operator(operator)*g
   d = -np.trace(g.imag)
   return d




def plot_central_dos(ht,energies=[0.0],num_rep=100,
                      mixing=0.7,eps=0.0001,green_guess=None,max_error=0.0001):
   """ Plots the density of states by using a 
    green function approach"""
   # get the dos
   dos = central_dos(ht,energies=energies,
             num_rep=num_rep,mixing=mixing,
             eps=eps,green_guess=green_guess,max_error=max_error)
   # plot the figure
   fig = py.figure() # create figure
   fig.set_facecolor("white") # face in white
   sp = fig.add_subplot(111) # create subplot
   sp.set_xlabel("Energy",size=20)
   sp.set_ylabel("Central DOS",size=20)
   sp.plot(energies,dos) # create the plot
   return fig




from .transporttk.landauer import landauer





def block2full(ht,sparse=False):
  """Convert a HTstructure with block diagonal Hamiltonian
  into the full form"""
  if not ht.block_diagonal: return ht # stop
  ho = ht.copy()
  ho.block_diagonal = False # set in false from now on
  nb = len(ht.central_intra) # number of blocks
  if nb>=2:
      lc = [csc_matrix(ht.central_intra[i][i].shape) for i in range(nb)]
      rc = [csc_matrix(ht.central_intra[i][i].shape) for i in range(nb)]
      lc[0] = csc_matrix(ht.left_coupling)
      rc[nb-1] = csc_matrix(ht.right_coupling)
  else: 
      print("Not implemented")
      raise # no central part
  # convert the central to sparse form
  central = [[None for i in range(nb)] for j in range(nb)]
  for i in range(nb):
    for j in range(nb):
      if ht.central_intra[i][j] is None: continue
      else:
        central[i][j] = csc_matrix(ht.central_intra[i][j])
  from scipy.sparse import vstack
  if sparse:
    ho.left_coupling = vstack(lc)
    ho.right_coupling = vstack(rc)
    ho.central_intra = bmat(ht.central_intra) # as sparse matrix
  else:
    ho.left_coupling = vstack(lc).todense()
    ho.right_coupling = vstack(rc).todense()
    ho.central_intra = bmat(central).todense() # as dense matrix
  return ho








def create_leads_and_central_list(h_right,h_left,list_h_central,
        right_coupling = None,left_coupling = None,
        scale_right_coupling = 1.0,
        scale_left_coupling = 1.0,
        coupling=lambda i,j: 1.0):
    """ Creates an HTjunction by giving the hamiltonian
       of the leads and the list of the center """
    # check the hamiltonians
  #  h_right.check()
  #  h_left.check()
    # convert to the classical way
    h_right = h_right.get_no_multicell()
    h_left = h_left.get_no_multicell()
    list_h_central = [h.get_no_multicell() for h in list_h_central]
    if len(list_h_central)==1: # only one central part
      return create_leads_and_central(h_right,h_left,list_h_central[0])
    ht = HTstructure(h_right) # create HTstructure
    ht.Hr = h_right.copy() # store
    ht.Hl = h_left.copy() # store
    # assign matrices of the leads
    ht.right_intra = h_right.intra.copy() 
    ht.right_inter = h_right.inter.copy()  
    ht.left_intra = h_left.intra.copy()  
    ht.left_inter = dagger(h_left.inter).copy() 
    # elecron hole stuff
    ht.has_eh = h_right.has_eh # if it has electron-hole
    ht.get_eh_sector = h_right.get_eh_sector # if it has electron-hole
    # create matrix of the central part and couplings to the leads
    from scipy.sparse import csc_matrix,bmat
    z = csc_matrix(h_right.intra*0.0j) # zero matrix
    num_central = len(list_h_central) # length of the central hamiltonian
    # create list of lists for the central part
    hc = [[None for i in range(num_central)] for j in range(num_central)]
    # create elements of the central hamiltonian
    for i in range(num_central):  # intra term of the central blocks
      hc[i][i] = list_h_central[i].intra.copy()  # intra term of the iesim
    for i in range(num_central-1):  # intra term of the central blocks
      tr = list_h_central[i].inter + list_h_central[i+1].inter
      tr = tr/2.*coupling(i,i+1)   # mean value of the hoppings
      hc[i][i+1] = tr # inter term of the iesim
      hc[i+1][i] = tr.H # inter term of the iesim
  
    # hoppings to the leads
    tcr = h_right.inter.copy()    # this is a matrix
    tcl = dagger(h_left.inter).copy()   # this is a matrix
    ht.block_diagonal = True
    # hoppings from the center to the leads
    if right_coupling is not None:
      ht.right_coupling = right_coupling*scale_right_coupling
    else:
      ht.right_coupling = h_right.inter*scale_right_coupling
    if left_coupling is not None:
      ht.left_coupling = left_coupling*scale_left_coupling
    else:
      ht.left_coupling = dagger(h_left.inter)*scale_left_coupling
  
    # assign central hamiltonian
    ht.central_intra = hc
#    # and modify the geometry of the central part
#    ht.central_geometry.supercell(num_central) 
#    # put if it si sparse
#    ht.is_sparse = list_h_central[0].is_sparse
    return ht


def eigenvalues(HT,numeig=10,effective=False,gf=None,full=False):
  """ Gets the lowest eigenvalues of the central part of the hamiltonian"""
  if not HT.block_diagonal:
    print(""" HTunction in eigenvalues must be block diagonal""")
    raise
  # if effective hamiltonian, just calculate the eigenvalues
  if effective: # effective hamiltonian
    print("Calculating eigenvalues of effective hamiltonian...")
    if HT.heff is None:  #if hasn't been calculated so far
      effective_central_hamiltonian(HT,write=False)
    heff = HT.heff # store the list of list with central EFF ham
    import scipy.sparse.linalg as lg
    eig,eigvec = lg.eigs(heff,k=numeig,which="LM",sigma=0.0)
    return eig

  from scipy.sparse import csc_matrix,bmat
  import scipy.sparse.linalg as lg
  if not effective: # do not use the effective hamiltonian, only the central
    intra = HT.central_intra # store the list of list with central ham
  numb = len(intra) # number of central blocks
  intrasp = [[None for i in range(numb)] for j in range(numb)]
  # assign onsite and couplings in sparse form
  if not HT.is_sparse:
    for i in range(numb):
      intrasp[i][i] = csc_matrix(intra[i][i])
    for i in range(numb-1):
      intrasp[i][i+1] = csc_matrix(intra[i][i+1])
      intrasp[i+1][i] = csc_matrix(intra[i+1][i])
  if HT.is_sparse:
    for i in range(numb):
      intrasp[i][i] = intra[i][i]
    for i in range(numb-1):
      intrasp[i][i+1] = intra[i][i+1]
      intrasp[i+1][i] = intra[i+1][i]
  intrasp = bmat(intrasp) # create sparse matrix
  if effective:
    evals,evecs = lg.eigs(intrasp,k=numeig,which="LM",sigma=0.0)
  if not effective:
    if full:  # full diagonalization
      from scipy.linalg import eigvalsh
      evals = eigvalsh(intrasp.todense())
    else:
      evals,evecs = lg.eigsh(intrasp,k=numeig,which="LM",sigma=0.0)
  return evals




def effective_central_hamiltonian(HT,energy=0.0,delta=0.0001,write=False):
   """ Plots the local density of states in the central part"""
   from .green import green_renormalization
   from .green import dyson
   from .hamiltonians import is_number
   # perform dyson calculation
   intra = HT.right_intra
   inter = HT.right_inter
#   gr = dyson(intra,inter,is_sparse=HT.is_sparse)
   ggg,gr = green_renormalization(intra,inter,energy=energy,delta=delta)
   HT.right_green = gr # save green function
   # left green function
   intra = HT.left_intra
   inter = HT.left_inter
#   gl = dyson(intra,inter,is_sparse=HT.is_sparse)
   ggg,gl = green_renormalization(intra,inter,energy=energy,delta=delta)
   HT.left_green = gl # save green function
   # save green functions
#   HT.write_green()
   # left selfenergy
   inter = HT.left_coupling
   selfl = inter@gl@dagger(inter) # left selfenergy
   # right selfenergy
   inter = HT.right_coupling
   selfr = inter@gr@dagger(inter) # right selfenergy
   # central green function
   intra = HT.central_intra
# dyson equation for the center     
   # full matrix
   if not HT.block_diagonal:
     heff = intra + selfl + selfr
     HT.heff = heff
     if save_heff:
       print("Saving effective hamiltonian in ",HT.file_heff)
       if write: # save hamiltonian if desired
         HT.write_heff() 
   # reduced matrix
   if HT.block_diagonal: 
     from copy import deepcopy
     heff = deepcopy(intra)
     heff[0][0] = intra[0][0] + selfl
     heff[-1][-1] = intra[-1][-1] + selfr
    # save the green function
     from scipy.sparse import bmat
#     HT.heff = bmat(heff)
     if write: # save hamiltonian
       print("Saving effective hamiltonian in ",HT.file_heff)
       HT.write_heff() 
   return heff




def get_tmatrix(ht,energy=0.0,delta=0.0001):
  """Calculate the S-matrix of an HTstructure"""
  if ht.block_diagonal: raise  # not implemented
  smatrix = get_smatrix(ht,energy=energy)
  return smatrix[0][1]






def get_surface_green(HT,energy=0.0,delta=1e-5):
   """Calculate left and right greeen functions"""
   from .green import green_renormalization
   # right lead
   intra = HT.right_intra
   inter = HT.right_inter
   ggg,gr = green_renormalization(intra,inter,energy=energy,delta=delta)
#   HT.right_green = gr # save green function
   # left lead
   intra = HT.left_intra
   inter = HT.left_inter
   ggg,gl = green_renormalization(intra,inter,energy=energy,delta=delta)
   return (gl,gr)




def get_bulk_green(HT,energy=0.0,delta=0.0001):
   """Calculate left and right bulk green functions"""
   from .green import green_renormalization
   # right lead
   intra = HT.right_intra
   inter = HT.right_inter
   gr,gr2 = green_renormalization(intra,inter,energy=energy,delta=delta)
#   HT.right_green = gr # save green function
   # left lead
   intra = HT.left_intra
   inter = HT.left_inter
   gl,gl2 = green_renormalization(intra,inter,energy=energy,delta=delta)
   return (gl,gr)









def get_central_selfenergies(HT,energy=0.0,delta=0.0001):
   """Calculate left and right selfenergies, coupled to central part"""
   (gl,gr) = get_surface_green(HT,energy=energy,delta=delta)
   inter = HT.left_coupling
   selfl = inter*gl*inter.H # left selfenergy
   # right selfenergy
   inter = HT.right_coupling
   selfr = inter*gr*inter.H # right selfenergy
   return (selfl,selfr) # return selfenergies


def get_surface_selfenergies(HT,energy=0.0,delta=0.0001,pristine=False):
   """Calculate left and right selfenergies"""
   (gl,gr) = get_surface_green(HT,energy=energy,delta=delta)
   if pristine:
     inter = HT.left_inter
   else:
     inter = HT.left_coupling
   selfl = inter@gl@np.conjugate(inter).T # left selfenergy
   # right selfenergy
   if pristine:
     inter = HT.right_inter
   else:
     inter = HT.right_coupling
   selfr = inter@gr@np.conjugate(inter).T # right selfenergy
   return (selfl,selfr) # return selfenergies




from .transporttk.builder import build

from .transporttk.smatrix import get_smatrix

from .transporttk.localprobe import LocalProbe



# function to work with spinless superconductivity
from numba import jit
from .. import algebra
import numpy as np

def onsite_delta_density_vev(h,nk=10,**kwargs):
    """Compute the expectation value of delta"""
    (es,ws) = h.get_eigenvectors(nk=nk,**kwargs) # compute the eigenvectors
    fac = 1./(nk**h.dimensionality) # number of kpoints
    wout = [] # empty list
    for i in range(len(es)):
        if es[i]<0.0: wout.append(ws[i]) # store
    wout = np.array(wout) # transform to array
    ni = h.intra.shape[0]//2 # number of sites
    p = np.zeros(ni,dtype=np.complex) # initialize
    d = np.zeros(ni,dtype=np.complex) # initialize
    p = compute_pairing(wout,p)*fac # compute the pairing
    d = compute_density(wout,d)*fac # compute the density
    return p,d # return the pairing and density

def onsite_delta_vev(h,T=0.0,nk=10,**kwargs):
    """Compute the expectation value of delta"""
    (es,ws) = h.get_eigenvectors(nk=nk,**kwargs) # compute the eigenvectors
    fac = 1./(nk**h.dimensionality) # number of kpoints
    wout = [] # empty list
    for i in range(len(es)):
        if T==0.0: # zero temperature
            if es[i]<0.0: wout.append(ws[i]) # store
        else: # finite temperature
            fd = 1./(np.exp(es[i]/T)+1.0) # occupation
            wout.append(ws[i]*fd)
    wout = np.array(wout) # transform to array
    ni = h.intra.shape[0]//2 # number of sites
    p = np.zeros(ni,dtype=np.complex) # initialize
    p = compute_pairing(wout,p)*fac # compute the pairing
    return p # return the pairing and density



@jit
def compute_pairing(ws,p):
    """Compute the pairing"""
    n = len(ws) # number of wavefunctions
    ni = len(ws[0])//2 # number of electron components
    for i in range(n): # loop over wavefunctions
        w = ws[i] # store
        for j in range(ni): # loop over components
            p[j] = p[j] + w[2*j]*np.conjugate(w[2*j+1]) # anomalous part
    return p # return expectation value


@jit
def compute_density(ws,p):
    """Compute the electronic density"""
    n = len(ws) # number of wavefunctions
    ni = len(ws[0])//2 # number of electron components
    for i in range(n): # loop over wavefunctions
        w = ws[i] # store
        for j in range(ni): # loop over components
            p[j] = p[j] + w[2*j]*np.conjugate(w[2*j]) # anomalous part
    return p # return expectation value










def swave_matrix(d):
    """Add swave pairing to a Hamiltonian, input is an array"""
    md = np.diag(d) # pairing matrix
    mdh = np.conjugate(md.T) # Hermitian
    m = algebra.bmat([[None,md],[mdh,None]]) # return matrix
    m = reorder(m) # reorder the entries
    return m


def add_swave_to_hamiltonian(h,d):
    """Add swave pairing to the hamiltonian"""
    if h.check_mode("spinless"):
      h.modify_hamiltonian_matrices(nambu) # modify the matrices
    elif h.check_mode("spinless_nambu"): pass # do nothing
    else: raise
    n = h.intra.shape[0]//2 # orbitals
    h.intra = h.intra + swave_matrix(np.ones(n)*d) # add matrix
    h.has_eh = True # has electron hole


def nambu(m):
    out = algebra.bmat([[m,None],[None,-np.conjugate(m)]])
    return reorder(out) # return matrix



def reorder(m):
  """Reorder a matrix that has electrons and holes"""
  R = np.array(np.zeros(m.shape)) # zero matrix
  nr = m.shape[0]//2 # number of positions
  for i in range(nr): # electrons
    R[i,2*i] = 1.0 # up electron
    R[i+nr,2*i+1] = 1.0 # down holes
  return R.T@m@R


def extract_swave(m):
    """Extract the swave vector form a matrix"""
    n = m.shape[0]//2 # number of orbitals
    out = np.zeros(n,dtype=np.complex) # initialize
    for i in range(n):
        out[i] = m[2*i,2*i+1]
    return out


def proje(n):
    """Return the projection operator in the electron sector"""
    m = np.zeros((n,n),dtype=np.complex)
    for i in range(n//2):
        m[2*i,2*i] = 1.0
    return m

def get_filling(h,nk=10,**kwargs):
    """Compute the expectation value of delta"""
    if not h.check_mode("spinless_nambu"): raise
    (es,ws) = h.get_eigenvectors(nk=nk,**kwargs) # compute the eigenvectors
    fac = 1./(nk**h.dimensionality) # number of kpoints
    wout = [] # empty list
    for i in range(len(es)):
        if es[i]<0.0: wout.append(ws[i]) # store
    wout = np.array(wout) # transform to array
    ni = h.intra.shape[0]//2 # number of sites
    d = np.zeros(ni,dtype=np.complex) # initialize
    d = compute_density(wout,d)*fac # compute the density
    return np.mean(d).real # return the pairing and density



import numpy as np
from scipy.sparse import csc_matrix,bmat,coo_matrix
from . import parallel


def collect_hopping(h):
    """Collect the hoppings in a new list"""
    td = dict() # dictionary
    zero = h.intra*0.0 # initialize
    for t in h.hopping: td[tuple(t.dir)] = zero.copy() # initialize
    for t in h.hopping: 
        td[tuple(t.dir)] += t.m # add
    ts = [] # empty list
    for key in td:
        if np.max(np.abs(td[key]))>1e-6:
          ts.append(Hopping(d=np.array(key),m=td[key]))

    return ts




class Hopping(): 
  def __init__(self,d=None,m=None):
    self.dir = np.array(d)
    self.m = m


def turn_multicell(h):
    """Transform a normal hamiltonian into a multicell hamiltonian"""
    if h.is_multicell: return h # if it is already multicell
    ho = h.copy() # copy hamiltonian
    # directions
    dirs = []
    if h.dimensionality == 0: ts = []
    elif h.dimensionality == 1: # one dimensional
      dirs.append(np.array([1,0,0]))
      ts = [h.inter.copy()]
      ho.inter = None
    elif h.dimensionality == 2: # two dimensional
      dirs.append(np.array([1,0,0]))
      dirs.append(np.array([0,1,0]))
      dirs.append(np.array([1,1,0]))
      dirs.append(np.array([1,-1,0]))
      ts = [h.tx.copy(),h.ty.copy(),h.txy.copy(),h.txmy.copy()]
      del ho.tx
      del ho.ty
      del ho.txy
      del ho.txmy
    else: raise
    dd = dict() # dictionary
    dd[(0,0,0)] = h.intra
    for (d,t) in zip(dirs,ts): 
        dd[tuple(d)] = t
        dd[tuple(-np.array(d))] = np.conjugate(t).T
    return set_dictionary(ho,dd) # return this Hamiltonian


def set_multihopping(ho,dd):
    """Set a multihopping as the Hamiltonian"""
    dd = dd.get_dict()
    return set_dictionary(ho,dd) # return this Hamiltonian



def set_dictionary(ho,dd):
    """Set a dictionary as the Hamiltonian"""
    ho.intra = dd[(0,0,0)] # intracell
    hoppings = [] # list of hoppings
    for d in dd: # loop over hoppings
        t = dd[d] # matrix
        if d==(0,0,0): continue
        hopping = Hopping() # create object
        hopping.m = t
        hopping.dir = np.array(d)
        hoppings.append(hopping) # store
#        hopping = Hopping() # create object
#        hopping.m = np.conjugate(t).T
#        hopping.dir = -np.array(d)
#        hoppings.append(hopping) # store
    ho.hopping = hoppings # store all the hoppings
    ho.is_multicell = True # multicell hamiltonian
    return ho


def generate_get_tij(h):
    """Get the hopping between cells with certain indexes"""
    hdict = h.get_multihopping().get_dict() # generate the multihopping object
    mzero = hdict[(0,0,0)]*0.
    def fun(rij=np.array([0.,0.,0.]),zero=False):
        drij = tuple([int(round(ir)) for ir in rij]) # put as integer
        if drij in hdict: return hdict[tuple(drij)]
        else:
            if zero: return mzero
            else: return None
    return fun

def hk_gen(h):
  """Generate a k dependent hamiltonian"""
  if not h.is_multicell:
      h = h.get_multicell()
  # get the non zero hoppings
  hopping = [] # empty list
  for t in h.hopping: # loop
    if h.is_sparse:
      if np.sum(np.abs(coo_matrix(t.m).data))>1e-7: hopping.append(t) # store this hopping
    else:
      if np.sum(np.abs(t.m))>1e-7: hopping.append(t) # store this hopping
  if h.dimensionality == 0: return lambda k: h.intra
  elif h.dimensionality == 1: # one dimensional
    def hk(k):
      """k dependent hamiltonian, k goes from 0 to 1"""
      mout = h.intra.copy() # intracell term
      for t in hopping: # loop over matrices
        tk = t.m * h.geometry.bloch_phase(t.dir,k) # k hopping
        mout = mout + tk 
      return mout
    return hk  # return the function
  elif h.dimensionality == 2: # two dimensional
    def hk(k):
      """k dependent hamiltonian, k goes from 0 to 1"""
      mout = h.intra.copy() # intracell term
      for t in hopping: # loop over matrices
        tk = t.m * h.geometry.bloch_phase(t.dir,k) # k hopping
        mout = mout + tk 
      return mout
    return hk  # return the function
  elif h.dimensionality == 3: # three dimensional
    def hk(k):
      """k dependent hamiltonian, k goes from 0 to 1"""
      mout = h.intra.copy() # intracell term
      for t in h.hopping: # loop over matrices
        tk = t.m * h.geometry.bloch_phase(t.dir,k) # k hopping
        mout = mout + tk 
      return mout
    return hk  # return the function
  else: raise







def turn_spinful(h,enforce_tr=False):
  """Turn a hamiltonian spinful"""
  from .increase_hilbert import spinful
  from .superconductivity import time_reversal
  if h.has_eh: raise
  if h.has_spin: return # return if already has spin
  h.has_spin = True # put spin
  def fun(m):
    return spinful(m)
  h.intra = fun(h.intra) # spinful intra
  for i in range(len(h.hopping)): 
    h.hopping[i].m = fun(h.hopping[i].m) # spinful hopping




def bulk2ribbon(hin,n=10,sparse=True,nxt=6,ncut=6):
  """ Create a ribbon hamiltonian object"""
  if not hin.is_multicell: h = turn_multicell(hin)
  else: h = hin # nothing othrwise
  hr = h.copy() # copy hamiltonian
  if sparse: hr.is_sparse = True # sparse output
  hr.dimensionality = 1 # reduce dimensionality
  # stuff about geometry
  hr.geometry = h.geometry.supercell((1,n)) # create supercell
  hr.geometry.dimensionality = 1
  hr.geometry.a1 = h.geometry.a1 # add the unit cell vector
  from . import sculpt # rotate the geometry
  hr.geometry = sculpt.rotate_a2b(hr.geometry,hr.geometry.a1,np.array([1.,0.,0.]))
  hr.geometry.celldis = hr.geometry.a1[0]
  get_tij = generate_get_tij(h) # return a function to abtain the hoppings
  def superhopping(dr=[0,0,0]): 
    """ Return a matrix with the hopping of the supercell"""
    intra = [[None for i in range(n)] for j in range(n)] # intracell term
    for ii in range(n): # loop over ii
      for jj in range(n): # loop over jj
        d = np.array([dr[0],ii-jj+dr[1],dr[2]])
        if d.dot(d)>ncut*ncut: continue # skip iteration
        m = get_tij(rij=d) # get the matrix
        if m is not None: intra[ii][jj] = csc_matrix(m) # store
        else: 
          if ii==jj: intra[ii][jj] = csc_matrix(h.intra*0.)
    intra = bmat(intra) # convert to matrix
    if not sparse: intra = intra.todense() # dense matrix
    return intra
  # get the intra matrix
  hr.intra = superhopping()
  # now do the same for the interterm
  hoppings = [] # list of hopings
  for i in range(-nxt,nxt+1): # loop over hoppings
    if i==0: continue # skip the 0
    d = np.array([i,0.,0.])
    hopp = Hopping() # create object
    hopp.m = superhopping(dr=d) # get hopping of the supercell
    hopp.dir = d
    hoppings.append(hopp)
  hr.hopping = hoppings # store the list
  hr.dimensionality = 1
  return hr 




def rotate90(h):
  """ Rotate 90 degrees the Hamiltonian"""
  ho = turn_multicell(h) # copy Hamiltonian
  hoppings = []
  for i in range(len(ho.hopping)):
    tdir = ho.hopping[i].dir 
    ho.hopping[i].dir = np.array([tdir[1],tdir[0],tdir[2]]) # new direction
  ho.geometry.a1,ho.geometry.a2 = h.geometry.a2,h.geometry.a1
  return ho


def rotate(h,m):
  """ Rotate the Hamiltonian"""
  ho = h.copy() # duplicate Hamiltonian
  m = np.matrix(m) # convert to matrix
  for i in range(len(h.hopping)):
    tdir = np.array(h.hopping[i].dir)*m.T # rotation of the direction
    tdir = [tdir[0,0],tdir[0,1],tdir[0,2]]
#    print(h.hopping[i].dir,tdir)
    ho.hopping[i].dir = tdir # new direction
  ho.hopping_dict = dict() # new dictionary
  ho.geometry.a1,ho.geometry.a2 = h.geometry.a2*m.T,ho.geometry.a1*m.T
  return ho


def basis_change(h,R):
  """Perform a change of basis in the Hamiltonian"""
  ho = h.copy() # duplicate Hamiltonian
  R = np.matrix(R) # convert to matrix
  for i in range(len(h.hopping)):
    m = h.hopping[i].m # rotation of the direction
    ho.hopping[i].m = R.H*m*R # Hamiltonian in new basis
  return ho





def clean(h,cutoff=0.0001):
  """Remove hoppings smaller than a certain quantity"""
  ho = h.copy() # copy hamiltonian
  raise 



def supercell_hamiltonian(hin,nsuper=[1,1,1],sparse=True,ncut=3,
                             **kwargs):
  """ Create a ribbon hamiltonian object"""
#  raise # there is something wrong with this function
#  print("This function might have something wrong")
  if not hin.is_multicell: h = turn_multicell(hin)
  else: h = hin # nothing otherwise
  hr = h.copy() # copy hamiltonian
  if sparse: hr.is_sparse = True # sparse output
  # stuff about geometry
  hr.geometry = h.geometry.get_supercell(nsuper,**kwargs) # create supercell
  n = nsuper[0]*nsuper[1]*nsuper[2] # number of cells in the supercell
  pos = [] # positions inside the supercell
  for i in range(nsuper[0]):
    for j in range(nsuper[1]):
      for k in range(nsuper[2]):
        pos.append(np.array([i,j,k])) # store position inside the supercell
  zero = csc_matrix(np.zeros(h.intra.shape,dtype=np.complex)) # zero matrix
  get_tij = generate_get_tij(h) # return a function to abtain the hoppings
  def superhopping(dr=[0,0,0]): 
    """ Return a matrix with the hopping of the supercell"""
    rs = [dr[0]*nsuper[0],dr[1]*nsuper[1],dr[2]*nsuper[2]] # supercell vector
    intra = [[None for i in range(n)] for j in range(n)] # intracell term
    for ii in range(n): intra[ii][ii] = zero.copy() # zero

    for ii in range(n): # loop over cells
      for jj in range(n): # loop over cells
        d = pos[jj] + np.array(rs) -pos[ii] # distance
      #  if d.dot(d)>ncut*ncut: continue # skip iteration
        m = get_tij(rij=d) # get the matrix
        if m is not None: 
          intra[ii][jj] = csc_matrix(m) # store
    intra = csc_matrix(bmat(intra)) # convert to matrix
    if not sparse: intra = intra.todense() # dense matrix
    return intra
  # get the intra matrix
  hr.intra = superhopping()
  # now do the same for the interterm
  hoppings = [] # list of hopings
  for i in range(-ncut,ncut+1): # loop over hoppings
    for j in range(-ncut,ncut+1): # loop over hoppings
      for k in range(-ncut,ncut+1): # loop over hoppings
        if i==j==k==0: continue # skip the intraterm
        dr = np.array([i,j,k]) # set as array
        hopp = Hopping() # create object
        hopp.m = superhopping(dr=dr) # get hopping of the supercell
        hopp.dir = dr
        if np.sum(np.abs(hopp.m))>0.00000001: # skip this matrix
          hoppings.append(hopp)
        else: pass
      hr.hopping = hoppings # store the list
  return hr 



from .current import derivative



def parametric_hopping_hamiltonian(h,cutoff=5,fc=None,rcut=5.0):
  """ Gets a first neighbor hamiltonian"""
  from .neighbor import parametric_hopping
  if fc is None:
    rcut = 2.1 # stop in this neighbor
    def fc(r1,r2):
      r = r1-r2
      r = r.dot(r)
      if 0.9<r<1.1: return 1.0
      else: return 0.0
  r = h.geometry.r    # x coordinate 
  g = h.geometry
  h.is_multicell = True 
# first neighbors hopping, all the matrices
  a1, a2, a3 = g.a1, g.a2, g.a3
  h.intra = h.spinless2full(parametric_hopping(r,r,fc)) # intra matrix
  # generate directions
  dirs = h.geometry.neighbor_directions(n=cutoff) # directions of the hoppings
  # generate hoppings
  h.hopping = [] # empty list
  for d in dirs: # loop over directions
        i1,i2,i3 = d[0],d[1],d[2] # extract indexes
        if i1==0 and i2==0 and i3==0: continue
        t = Hopping() # hopping class
        da = a1*i1+a2*i2+a3*i3 # direction
        r2 = [ri + da for ri in r]
        if not close_enough(r,r2,rcut=rcut): # check if we can skip this one
#          print("Skipping hopping",[i1,i2,i3])
          continue
        t.m = h.spinless2full(parametric_hopping(r,r2,fc))
        t.dir = [i1,i2,i3] # store direction
        if np.sum(np.abs(t.m))>0.00001: h.hopping.append(t) # append 
  return h







def parametric_matrix(h,cutoff=2,fm=None):
  """ Gets a first neighbor hamiltonian"""
  from .neighbor import parametric_hopping
  if fm is None: raise
  r = h.geometry.r    # x coordinate 
  g = h.geometry
  h.is_multicell = True
# first neighbors hopping, all the matrices
  a1, a2, a3 = g.a1, g.a2, g.a3
  h.intra = h.spinless2full(fm(r,r)) # intra matrix
  # generate directions
  dirs = h.geometry.neighbor_directions(n=cutoff) # directions of the hoppings
  # generate hoppings
  h.hopping = [] # empty list
  def gett(d):
        i1,i2,i3 = d[0],d[1],d[2] # extract indexes
        if i1==0 and i2==0 and i3==0: return None
        t = Hopping() # hopping class
        da = a1*i1+a2*i2+a3*i3 # direction
        r2 = [ri + da for ri in r]
        t.m = h.spinless2full(fm(r,r2))
        t.dir = [i1,i2,i3] # store direction
        if np.sum(np.abs(t.m))>0.00001: return t
        else: return None
#  from . import parallel
  ts = parallel.pcall(gett,dirs) # get hoppings
#  ts = [gett(d) for d in dirs] # get hoppings in serie
  ts = [x for x in ts if x is not None] # remove Nones
  h.hopping = ts # store
  return h



# rename this function
first_neighbors = parametric_hopping_hamiltonian

def read_from_file(input_file="hamiltonian.wan"):
  """Generate a function that return hopping matrices"""
  mt = np.genfromtxt(input_file) # get file
  m = mt.transpose() # transpose matrix
  nmax = int(np.max([np.max(m[i])for i in range(3)]))
  ncells = [nmax,nmax,nmax] # number of cells
  # read the hamiltonian matrices
  tlist = []
  norb = np.max([np.max(np.abs(m[3])),np.max(np.abs(m[4]))])
  norb = int(norb)
  zeros = np.matrix(np.zeros((norb,norb),dtype=np.complex)) # zero matrix
  def get_t(i,j,k):
    mo = zeros.copy() # copy matrix
    found = False
    for l in mt: # look into the file
      if i==int(l[0]) and j==int(l[1]) and k==int(l[2]): # right hopping
        mo[int(l[3])-1,int(l[4])-1] = l[5] + 1j*l[6] # store element
        found  = True # matrix has been found
    if found:  return mo # return the matrix
    else: return None # return nothing if not found
  tdict = dict() # dictionary
  for i in range(-ncells[0],ncells[0]+1):
    for j in range(-ncells[1],ncells[1]+1):
      for k in range(-ncells[2],ncells[2]+1):
        matrix = get_t(i,j,k) # read the matrix
        if matrix is None: continue # skip if matrix not found
        tdict[(i,j,k)] = matrix # store matrix
  # now create the function
  def get_hopping(i,j,k):
    """Get hopping in a certain direction"""
    try: # look into the dictionary
      return tdict[(i,j,k)]
    except: return zeros  # return zero matrix if not found
  # check that the function returns a Hermitian Hamiltonian
  for i in range(-ncells[0],ncells[0]+1):
    for j in range(-ncells[1],ncells[1]+1):
      for k in range(-ncells[2],ncells[2]+1):
        dm = get_hopping(i,j,k) - get_hopping(-i,-j,-k).H
        if np.sum(np.abs(dm))>0.0001: raise
  print("Hopping generator is Hermitian")
  return get_hopping # return the function
     

def save_multicell(ns,ms,output_file="multicell.sym"):
  """Save a multicell Hamiltonian in a file"""
  fo = open(output_file,"w") # open file
  for (n,m) in zip(ns,ms): # loop over hoppings
    for i in range(m.shape[0]):
      for j in range(m.shape[0]):
        fo.write(str(n[0])+"    ") # cell number
        fo.write(str(n[1])+"    ") # cell number
        fo.write(str(n[2])+"    ") # cell number
        fo.write(str(i+1)+"    ") # index
        fo.write(str(j+1)+"    ") # index
        fo.write(str(m[i,j].real)+"    ") # index
        fo.write(str(m[i,j].imag)+"\n") # index
  fo.close() # close file



def pairs2hopping(ps):
  """Takes as input a list of pairs vector/matrix, return a list of classess"""
  hopping = [Hopping(d,m) for (d,m) in ps] # empty list
  return hopping



def close_enough(rs1,rs2,rcut=2.0):
  """CHeck if two sets of positions are at a distance
  at least rcut"""
  rcut2 = rcut*rcut # square of the distance
  for ri in rs1:
    for rj in rs2:
      dr = ri - rj # vector
      dr = dr.dot(dr) # distance
      if dr<rcut2: return True
  return False



def turn_no_multicell(h):
  """Converts a Hamiltonian into the non multicell form"""
  if not h.is_multicell: return h # Hamiltonian is already fine
  ho = h.copy() # copy Hamiltonian
  ho.is_multicell = False
  if ho.dimensionality==0: pass
  elif ho.dimensionality==1:
    ho.inter = ho.intra*0.
  elif ho.dimensionality==2:
    ho.tx = ho.intra*0.
    ho.txy = ho.intra*0.
    ho.txmy = ho.intra*0.
    ho.ty = ho.intra*0.
  else: raise
  for t in h.hopping: # loop over hoppings
    if h.dimensionality==0: pass # one dimensional
    elif h.dimensionality==1: # one dimensional
      if t.dir[0]==1 and t.dir[1]==0 and t.dir[2]==0: # 
        ho.inter = t.m # store
      elif np.sum(np.abs(t.m))>0.0001 and np.max(np.abs(t.dir))>1: raise # Uppps, not possible
    elif h.dimensionality==2: # two dimensional
      if t.dir[0]==1 and t.dir[1]==0 and t.dir[2]==0: # 
        ho.tx = t.m # store
      elif t.dir[0]==0 and t.dir[1]==1 and t.dir[2]==0: # 
        ho.ty = t.m # store
      elif t.dir[0]==1 and t.dir[1]==1 and t.dir[2]==0: # 
        ho.txy = t.m # store
      elif t.dir[0]==1 and t.dir[1]==-1 and t.dir[2]==0: # 
        ho.txmy = t.m # store
      elif np.sum(np.abs(t.m))>0.0001 and np.max(np.abs(t.dir))>1: raise # Uppps, not possible
    else: raise
  ho.hopping = [] # empty list
  return ho



def kchain(h,k=[0.,0.,0.]):
  """Return the onsite and hopping for a particular k"""
  if not h.is_multicell: h = h.get_multicell()
  dim = h.dimensionality # dimensionality
  if dim==1: # 1D
      for t in h.hopping:
          if t.dir[0]==1: return h.intra,t.m
      raise
  elif dim>1: # 2D or 3D
    intra = np.zeros(h.intra.shape) # zero amtrix
    inter = np.zeros(h.intra.shape) # zero amtrix
    intra = h.intra # initialize
    for t in h.hopping: # loop over hoppings
      tk = t.m * h.geometry.bloch_phase(t.dir,k) # k hopping
      if t.dir[dim-1]==0: intra = intra + tk # add contribution 
      if t.dir[dim-1]==1: inter = inter + tk # add contribution 
    return intra,inter 
  else: raise




def get_hopping_dict(h):
    """Return the hopping dictionary"""
    h = h.get_multicell()
    out = dict()
    out[(0,0,0)] = h.intra
    for t in h.hopping: 
        if np.max(np.abs(t.m))>1e-7:
          out[tuple(t.dir)] = t.m # store
    return out # return


#
#def multiply_hopping_dict(hop1,hop2):
#    """Multiply two hopping dictionaries"""
#    out = dict() # create dictionary
#    for key1 in hop1:
#        for key2 in hop2:
#            key = tuple(np.array(key1) + np.array(key2))
#            m = hop1[key1]@hop2[key2] # multiply
#            if np.max(np.abs(m))>1e-6: # if non-zero
#                if key in out: out[key] = out[key] + m # add
#                else: out[key] = m # store
#    return out # return output
#
#
#
#def add_hopping_dict(hop1,hop2):
#    """Multiply two hopping dictionaries"""
#    out = dict() # create dictionary
#    keys = [key1 for key1 in hop1]
#    for key2 in hop2:
#        if key2 not in keys: keys.append(key2) # store
#    for key in keys:
#        m = 0 # initialize
#        if key in hop1: m = m + hop1[key]
#        if key in hop2: m = m + hop2[key]
#        out[key] = m # store
#    return out
#
#
#
#class MultiHopping():
#    """Class for a multihopping"""
#    def __init__(self,a=dict()):
#        if type(a)==dict: # dictionary type
#            self.dict = a # dictionary
#        else: return NotImplemented
#    def __add__(self,a):
#        if type(a)!=MultiHopping: return NotImplemented
#        out = MultiHopping() # create a new object
#        out.dict = add_hopping_dict(self.dict,a.dict)
#        return out
#    def __mul__(self,a):
#        from .hamiltonians import is_number
#        if type(a)==MultiHopping: 
#            out = MultiHopping() # create a new object
#            out.dict = multiply_hopping_dict(self.dict,a.dict)
#            return out
#        elif is_number(a):
#            out = MultiHopping() # create a new object
#            for key in self.dict:
#                out.dict[key] = a*self.dict[key] 
#            return out
#        else: return NotImplemented
#    def __rmul__(self,a): return self*a
#    def __neg__(self): return (-1)*self
#    def __sub__(self,a): return self + (-self)
#    def get_dict(self):
#        return self.dict # dictionary
#    def get_dagger(self):
#        out = MultiHopping() # create a new object
#        for key in self.dict:
#            out.dict[key] = np.conjugate(self.dict[key]).T
#        return out
#
#
#
#
#

from .multihopping import MultiHopping




import numpy as np

def orthogonality_distribution(vs,n=10):
    """Given a set of eigenvectors, compute the distribution of their
    orthogonality"""
    n0 = len(vs)
    m = [] # output
    for i in range(n0):
        mi = 1.0
        for j in range(n0):
            if i==j: continue
            d = 1.-np.abs(vs[i].dot(np.conjugate(vs[j])))**2 # overlap
            mi *= d
        m.append(mi) # store
    out = np.histogram(m,bins=n,range=(0.,1.)) # return histogram
    dd = out[0]
    dd = dd/np.sum(dd)
    return out[1][0:n],dd





# module to handle different types of data 
import matplotlib.pyplot as plt



class Band():
  """ Data of a band structure"""
  def __init__(self):
    self.dimension = 1
    self.x = []
    self.y = []
    self.z = None
  def plot(self):
    fig = plt.figure() # new figure
    fig.set_facecolor("white") # panel in white
    fig.scatter(self.x,self.y) # plot the data
    return fig
  def show(self):
    fig = self.plot() # get the figure
    plt.show() # show figure






from scipy.sparse import issparse,bmat
from scipy.sparse import csc_matrix as csc
from scipy.sparse import csc_matrix
import scipy.linalg as dlg
import scipy.sparse.linalg as slg
import numpy as np
#from .algebratk import sparsetensor
from numba import jit
from . import parallel




import numbers
def isnumber(s):
    return isinstance(s, numbers.Number)



maxsize = 10000

def ismatrix(m):
    return type(m)==np.ndarray or issparse(m) or type(m)==np.matrix



def hermitian(m):
    return np.transpose(np.conjugate(m))

H = hermitian # alias
get_dagger = hermitian
dagger = hermitian

def inv(m):
    return dlg.inv(todense(m))


def trace(m):
    return np.trace(m)


def densebmat(m):
    """Turn a block matrix dense"""
    ms = [[todense(mi) for mi in mij] for mij in m]
    return todense(bmat(ms)) # return block matrix

def direct_sum(ms):
    mout = [[None for i in range(len(ms))] for j in range(len(ms))]
    for i in range(len(ms)): mout[i][i] = ms[i]
    return densebmat(mout)


def dot(a,b):
    """Compute the scalar product"""
    return np.dot(np.conjugate(a),b)

def braket_wAw(w,A,wi=None):
  """
  Compute the braket of a wavefunction
  """
  if wi is None: wi = w
  if issparse(A): # sparse matrices
    return (np.conjugate(wi)@A@w) # modern way
  else: # matrices and arrays
    return (np.conjugate(wi)@np.array(A)@w) # modern way


def todense(m):
    """Turn a matrix dense"""
    if issparse(m):
        if m.shape[0]<maxsize: return np.array(m.todense())
        else: raise
    else: return m


def braket_ww(w,wi):
  """
  Compute the braket of two wavefunctions
  """
  w = matrix2vector(w) # convert to vector
  wi = matrix2vector(wi) # convert to vector
  return (np.conjugate(w)@wi) # modern way




def disentangle_manifold(wfs,A):
  """
  Disentangles the wavefunctions of a degenerate manifold
  by expressing them in terms of eigenvalues of an input operator
  """
  ma = get_representation(wfs,A) # get the matrix form of the operator
  wfsout = [] # empty list
  evals,evecs = dlg.eigh(ma) # diagonalize

  evecs = evecs.transpose() # transpose eigenvectors
  for v in evecs: # loop over eigenvectors
    wf = wfs[0]*0.0j
    for (i,iv) in zip(range(len(v)),v): # loop over components
      wf += iv*wfs[i] # add contribution
    wfsout.append(wf.copy()) # store wavefunction
  return wfsout



def get_representation(wfs,A):
  """
  Gets the matrix representation of a certain operator
  """
  n = len(wfs) # number of eigenfunctions
  ma = np.zeros((n,n),dtype=np.complex) # representation of A
  sa = csc(A) # sparse matrix
  for i in range(n):
    vi = csc(np.conjugate(wfs[i])) # first wavefunction
    for j in range(n):
      vj = csc(wfs[j]).transpose() # second wavefunction
      data = (vi*sa*vj).todense()[0,0]
      ma[i,j] = data
  return ma





## routines for diagonalization ##

error = 1e-7

def todouble(vs,ind):
    """Double the eigenvectors"""
    nv = vs.shape[0]
    dim = vs.shape[1]
    vout = np.zeros((dim*2,nv),dtype=np.complex) # output vector
    return todouble_jit(vs,ind,vout,nv,dim)

@jit(nopython=True)
def todouble_jit(vs,ind,vout,nv,dim):
    """Double the eigenvectors, jit routine"""
    for i in range(dim):
        vout[2*i+ind,:] = vs[i,:]
    return vout



accelerate = False 

def eigh(m):
    """Wrapper for linalg"""
    m = todense(m)
    if np.max(np.abs(m.imag))<error: m = m.real # real matrix
    if not accelerate: return dlg.eigh(m)
#    from . import algebraf90
    # check if doing slices helps
    n = m.shape[0] # size of the matrix
    mo = m[0:n:2,1:n:2] # off diagonal is zero
#    if False: # assume block diagonal
    if np.max(np.abs(mo))<error: # assume block diagonal
        # detected block diagonal
        ms = [m[0:n:2,0:n:2],m[1:n:2,1:n:2]]
        out = parallel.pcall(lambda x: dlg.eigh(x),ms)
        (es0,vs0) = out[0]
        (es1,vs1) = out[1]
   #     (es0,vs0) = eigh(m[0:n:2,0:n:2]) # recall
   #     (es1,vs1) = eigh(m[1:n:2,1:n:2]) # recall
        es = np.concatenate([es0,es1]) # concatenate array
        #vs0 = algebraf90.todouble(vs0.T,0)
        #vs1 = algebraf90.todouble(vs1.T,1)
        vs0 = todouble(vs0,0) # double the degrees of freedom
        vs1 = todouble(vs1,1) # double the degrees of freedom
        vs = np.concatenate([vs0.T,vs1.T]).T
        return (es,vs) # return the eigenvaleus and eigenvectors

    else:
      if np.max(np.abs(m.imag))<error: # assume real
          return dlg.eigh(m.real) # diagonalize real matrix
      else: return dlg.eigh(m) # diagonalize complex matrix


def eigvalsh(m):
    """Wrapper for linalg"""
    m = todense(m) # turn the matrix dense
    if np.max(np.abs(m.imag))<error: m = m.real # real matrix
    if not accelerate: return dlg.eigvalsh(m)
    # check if doing slices helps
    n = m.shape[0] # size of the matrix
    mo = m[0:n:2,1:n:2] # off diagonal is zero
#    if False: # assume block diagonal
    if np.max(np.abs(mo))<error: # assume block diagonal
        # detected block diagonal
        es0 = dlg.eigvalsh(m[0:n:2,0:n:2]) # recall
        es1 = dlg.eigvalsh(m[1:n:2,1:n:2]) # recall
        es = np.concatenate([es0,es1]) # concatenate array
        return es

    else:
      if np.max(np.abs(m.imag))<error: # assume real
          return dlg.eigvalsh(m.real) # diagonalize real matrix
      else: return dlg.eigvalsh(m) # diagonalize complex matrix



def matrix2vector(v):
    """Transform a matrix into a vector"""
    if issparse(v): # sparse matrix
      v = v.todense() # convert to conventional matrix
    v = np.array(v) # convert to array
    if len(v.shape)==1: return v
    else: return v.reshape(v.shape[0]*v.shape[1])




def smalleig(m,numw=10,evecs=False,tol=1e-7):
    """
    Return the smallest eigenvalues using arpack
    """
    m = csc_matrix(m) # sparse matrix
    try:
        eig,eigvec = slg.eigsh(m,k=numw,which="LM",sigma=0.0,
                                        tol=tol)
        if evecs:  return eig,eigvec.transpose()  # return eigenvectors
        else:  return eig  # return eigenvalues
    except:
        print("Switch to dense")
        if m.shape[0]>maxsize: raise
        else:
            if not evecs: return eigvalsh(todense(m))
            else: return eigh(todense(m))



def spectral_gap(m,numw=10,**kwargs):
    """
    Compute the spectral gap
    """
    es = smalleig(m,numw=10,evecs=False,**kwargs)
    ev = es[es<0.]
    ec = es[es>0.]
    if len(ev)==0 or len(ec)==0:
        if numw<100: return gap(m,numw=2*numw,**kwargs)
        else: raise
    g = np.min(np.abs(ev))+np.min(ec) # gap
    return g # return gap







def sqrtm(M):
    """Square root for Hermitian matrix"""
    (m2,evecs) = sqrtm_rotated(M)
    m2 = dagger(evecs) @ m2 @ evecs  # change of basis
    return m2 # return matrix



def sqrtm_rotated(M,positive=True):
    """Square root for Hermitian matrix in the diagonal basis,
    and rotation matrix"""
    M = (M + dagger(M))/2. # make Hermitian
    (evals,evecs) = dlg.eigh(M) # eigenvals and eigenvecs
    if positive:
        if np.min(evals)<0.:
            print("Matrix is not positive defined")
            evals[evals<0.] = 1e-7
    evecs = dagger(np.matrix(evecs)) # change of basis
    m2 = np.matrix([[0.0j for i in evals] for j in evals]) # create matrix
    for i in range(len(evals)):
        m2[i,i] = np.sqrt(np.abs(evals[i])) # square root
    return (m2,evecs) # return matrix






from __future__ import print_function
from scipy.interpolate import interp1d
import numpy as np

# perform interpolation a matrix 
def intermatrix(fin,xs=np.linspace(-5.0,5.0,20)):
  """Return a function capable to interpolate the
  input function between values in the interval. Values
  outside the intervale will be returned as zero"""
  m = fin(0) # call once
  ar = np.zeros((len(xs),m.shape[0],m.shape[1])) # empty array, real part
  ai = np.zeros((len(xs),m.shape[0],m.shape[1])) # empty array, imag part
  for i in range(len(xs)): # loop
    m = fin(xs[i]) # call the function
    ar[i,:,:] = m.real # real part
    ai[i,:,:] = m.imag # imaginary part
  zero = np.matrix(np.zeros(m.shape,dtype=np.complex)) # zero matrix
  fr = interp1d(xs, ar, axis=0,kind=3,fill_value=zero)
  fi = interp1d(xs, ai, axis=0,kind=3,fill_value=zero)
  def fout(e): # output function
    return np.matrix(fr(e) + 1j*fi(e)) # return 
  return fout # return function


def interpolator2d(x,y,z):
  """Return a 2D interpolator"""
  x = np.array(x)
  y = np.array(y)
  from scipy.interpolate import interp2d
  ny = (y.max()-y.min())/abs(y[1]-y[0])
  ny = int(round(ny)) + 1
  nx = len(z)/ny
  nx = int(nx)
  ny = int(ny)
  x0 = np.linspace(min(x),max(x),nx)
  y0 = np.linspace(min(y),max(y),ny)
  xx, yy = np.meshgrid(x0, y0)
  Z = np.array(z).reshape(nx,ny) # makes a (Zy,Zx) matrix out of z
  T = Z.T
  f = interp2d(x0, y0, T, kind='linear')
  return f



def interpolator2d(x,y,z,mode=None):
    from scipy.interpolate import griddata
    from scipy.interpolate import NearestNDInterpolator
    def f(p):
        return griddata((x,y), z,p[:,0:2], method='nearest')
    if mode is None: return f
    elif mode=="periodic":
        def f0(k):
            """Define a periodic function"""
            k = k[:,0:2]%1.
            return f(k)
        return f0
    else: raise




def periodic_grid2mesh(ds,qs):
    """Given a certain data in a (periodic) grid, return in the input mesh"""
    nx = ds.shape[0]
    ny = ds.shape[1]
    grid_kx, grid_ky = np.mgrid[0:1:nx*1j, 0:1:ny*1j] # kx and ky
    ksg = np.array([grid_kx, grid_ky]).reshape((2,nx*ny)).T # create points
    fo = interpolator2d(ksg[:,0],ksg[:,1],ds.reshape((nx*ny)),mode="periodic")
    out = fo(qs[:,0:2])
    return out



from .interpolatetk.atomicinterpolation import atomic_interpolation





from . import multicell
import numpy as np

def slab(h,nz=1,nx=1,ny=1,sparse=False,ncut=3):
  """Create slab"""
  hout = multicell.supercell(h,nsuper=[nx,ny,nz],sparse=sparse,ncut=ncut)
  hout.dimensionality = 2
  hout.geometry.dimensionality = 2
  hopping = [] # empty list
  for t in hout.hopping:
    d = t.dir
    if np.abs(d[2])<0.1: # not in the z direction
      hopping.append(t)
  hout.hopping = hopping
  return hout




from .rg import green_renormalization
from .. import algebra

def green_kchain(h,k=0.,energy=0.,delta=0.01,only_bulk=True,
                    error=0.0001,hs=None,**kwargs):
  """ Calculates the green function of a kdependent chain for a 2d system """
  def gr(ons,hop):
    """ Calculates G by renormalization"""
    gf,sf = green_renormalization(ons,hop,energy=energy,nite=None,
                            error=error,info=False,delta=delta)
    if hs is not None: # surface matrix provided
      ez = (energy+1j*delta)*np.identity(h.intra.shape[0]) # energy
      sigma = hop@sf@algebra.dagger(hop) # selfenergy
      if callable(hs): ons2 = hs(k)
      else: ons2 = hs
      sf = algebra.inv(ez - ons2 - sigma) # return Dyson
    if only_bulk:  return gf
    else:  return gf,sf
  (ons,hop) = get1dhamiltonian(h,k,**kwargs) # get 1D Hamiltonian
  return gr(ons,hop)  # return green function




def get1dhamiltonian(hin,k=[0.0,0.,0.],reverse=False,**kwargs):
  """Return onsite and hopping matrix for a 1D Hamiltonian"""
  from .. import multicell
  (ons,hop) = multicell.kchain(hin,k=k,**kwargs)
  if reverse: return (ons,algebra.dagger(hop)) # return 
  else: return (ons,hop) # return 




def green_kchain_evaluator(h,k=0.,delta=0.01,only_bulk=True,
                    error=0.0001,hs=None,reverse=False):
  """ Calculates the green function of a kdependent chain for a 2d system """
  def gr(ons,hop,energy):
    """ Calculates G by renormalization"""
    gf,sf = green_renormalization(ons,hop,energy=energy,nite=None,
                            error=error,info=False,delta=delta)
    if hs is not None: # surface matrix provided
      ez = (energy+1j*delta)*np.identity(h.intra.shape[0]) # energy
      sigma = hop@sf@algebra.dagger(hop) # selfenergy
      if callable(hs): ons2 = ons + hs(k)
      else: ons2 = ons + hs
      sf = algebra.inv(ez - ons2 - sigma) # return Dyson
    # which green function to return
    if only_bulk:  return gf
    else:  return gf,sf
  (ons,hop) = get1dhamiltonian(h,k,reverse=reverse) # get 1D Hamiltonian
  def fun(energy): # evaluator
    return gr(ons,hop,energy)  # return green function
  return fun # return the function





from __future__ import print_function,division
import numpy as np
from . import geometry
from . import sculpt

def twisted_bilayer(m0,**kwargs):
    """Geomtry of a twisted bilayer"""
    return twisted_multilayer(m0=m0,rot=[1,0],**kwargs)

#def twisted_bilayer(m0=3,rotate=True,shift=[0.,0.],center="AB/BA",
#  sublattice=True,r=1,g=None,dz=1.5):
#  """Return the geometry for twisted bilayer graphene"""
#  if g is None: g = geometry.honeycomb_lattice()
#  g.has_sublattice = False
#  if sublattice: # trace the sublattice using a dirty trick
#    g.z[0] += 0.001
#    g.z[1] -= 0.001
#    g.xyz2r()
#  else: pass
#  g = geometry.non_orthogonal_supercell(g,m=[[-1,0,0],[0,1,0],[0,0,1]])
##  m0 = 3
##  r = 1
#  theta = np.arccos((3.*m0**2+3*m0*r+r**2/2.)/(3.*m0**2+3*m0*r+r**2))
#  print("Theta",theta*180.0/np.pi)
#  g.angle = theta # terrible workaround
#  g.data["angle"] = theta # store the angle of the geometry
#  nsuper = [[m0,m0+r,0],[-m0-r,2*m0+r,0],[0,0,1]]
#  g = geometry.non_orthogonal_supercell(g,m=nsuper,
#           reducef=lambda x: 3*np.sqrt(x))
#  g1 = g.copy()
#  g1.shift([1.,0.,0.]) 
#  g.z -= dz
#  g.xyz2r() # update
#  if rotate: # rotate one of the layers
#    g1 = g1.rotate(theta*180/np.pi)
#    g1s = g1.supercell(2) # supercell
#    g1s.z += dz
#    g1s.x += shift[0] # shift upper layer
#    g1s.y += shift[1] # shift upper layer
#    g1s.xyz2r() # update
#    g1.a1 = g.a1
#    g1.a2 = g.a2
#    rs = sculpt.retain_unit_cell(g1s.r,g.a1,g.a2,g.a3,dim=2) # get positions
#  else: # do not rotate
#    rs = np.array(g1.r) # same coordinates
#    rs[:,2] = 1.0 # set as one
#  g1.r = np.concatenate([rs,g.r])
# 
#  g1.r2xyz() # update
#  g1.real2fractional() # update fractional coordinates 
#  if center=="AB/BA": pass # do nothing 
#  elif center=="AA": 
#    g1.frac_x = (g1.frac_x)%1 # to the unit cell
#    g1.fractional2real()
#  elif center=="AB": 
#    g1.frac_x = (g1.frac_x)%1 # to the unit cell
#    g1.frac_y = (g1.frac_y)%1 # to the unit cell
#    g1.frac_y += 0.5  # to the unit cell
#    g1.frac_x = (g1.frac_x)%1 # to the unit cell
#    g1.frac_y = (g1.frac_y)%1 # to the unit cell
#    g1.frac_x -= 1./3.
#    g1.frac_y -= 1./3.
#    g1.frac_x = (g1.frac_x)%1 # to the unit cell
#    g1.frac_y = (g1.frac_y)%1 # to the unit cell
#    g1.fractional2real()
#  else: raise
#  if sublattice: # recover the sublattice
#    g1.has_sublattice = True # say that it has
#    sl = []
#    for r in g1.r: # loop over positions
#      if np.abs(r[2]-1.5)<0.01: # upper layer
#        if r[2]-1.5>0.0: sl.append(-1.) # A sublattice
#        else: sl.append(1.) # B sublattice
#      elif np.abs(r[2]+1.5)<0.01: # lower layer
#        if r[2]+1.5>0.0: sl.append(-1.) # A sublattice
#        else: sl.append(1.) # B sublattice
#    g1.z = np.round(g1.z,2) # delete the small shift
#    g1.xyz2r() # update coordinates
#    g1.sublattice = np.array(sl) # store sublattice
#  g1 = sculpt.rotate_a2b(g1,g1.a1,np.array([1.,0.,0.])) # rotate
#  g1.get_fractional() # get fractional coordinates 
#  return g1


def twisted_supercell(g,m0=3,r=1):
  """Return the supercell for a twisted system"""
  g.has_sublattice = False # no sublattice
  g = geometry.non_orthogonal_supercell(g,m=[[-1,0,0],[0,1,0],[0,0,1]])
  theta = np.arccos((3.*m0**2+3*m0*r+r**2/2.)/(3.*m0**2+3*m0*r+r**2))
  print("Theta",theta*180.0/np.pi)
  nsuper = [[m0,m0+r,0],[-m0-r,2*m0+r,0],[0,0,1]]
  g = geometry.non_orthogonal_supercell(g,m=nsuper,
           reducef=lambda x: 3*np.sqrt(x))
  return g




def twisted_multilayer(m0=3,rotate=True,shift=None,
  sublattice=True,r=1,rot=[1,1,0,0],g=None,dz=3.0):
  """Return the geometry for twisted multilayer graphene"""
  if g is None: g = geometry.honeycomb_lattice() # default is honeycomb
  else: g = geometry.get_geometry(g) # get the geometry
#  g.has_sublattice = False # no sublattice
  from . import supercell as supercelltk
# put the right angle between lattice vectors
  g = supercelltk.target_angle_volume(g,angle=1./3.,volume=1)
# old version
#  g = geometry.non_orthogonal_supercell(g,m=[[-1,0,0],[0,1,0],[0,0,1]])
#  print(g.a1,g.a2,g.a1.dot(g.a2)) ; exit()
  g0 = g.copy() # copy geometry
  theta = np.arccos((3.*m0**2+3*m0*r+r**2/2.)/(3.*m0**2+3*m0*r+r**2))
  if shift is None:
      shift = [[0.,0.] for r in rot] # initialize
  print("Theta",theta*180.0/np.pi)
  nsuper = [[m0,m0+r,0],[-m0-r,2*m0+r,0],[0,0,1]]
  g = geometry.non_orthogonal_supercell(g,m=nsuper,
           reducef=lambda x: 3*np.sqrt(x))
  if rotate: # rotate one of the layers
    gs = [] # empty list with geometries
    ii = 0
    for i in rot: # loop
        if i!=0 and i!=1: raise # nope
        dr = shift[ii][0]*g0.a1 + shift[ii][1]*g0.a2 # shift geometry
        dr = dr + np.array([0.,0.,dz*(ii-len(rot)/2.+.5)]) # shift layer
        gs.append(rotate_layer(g,i*theta,dr=dr))
      #  gs.append(g.copy())
        ii += 1 # increase counter
  else: raise
#  g.r = np.concatenate([g1.r,g.r,g2.r]).copy()
  g.r = np.concatenate([gi.r for gi in gs]).copy() # all the positions
  if g.has_sublattice:
    g.sublattice = np.concatenate([gi.sublattice for gi in gs]).copy() 
#  g.r = np.concatenate([g2.r,g.r]).copy()
 # g.r = g1.r
  g.r2xyz() # update
  g.real2fractional() # update fractional coordinates 
  g = sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.])) # rotate
  g.get_fractional() # get fractional coordinates 
  return g









def twisted_trilayer(m0=3):
  return twisted_multilayer(m0=m0,shift=[-1.,0.,1.])


def generalized_twisted_multilayer(m0=3,rotate=True,shift=[0.,0.],
  sublattice=True,r=1,rot=[1,1,0,0],gf=None,dz=3.0):
  """Return the geometry for twisted bilayer graphene"""
  theta = np.arccos((3.*m0**2+3*m0*r+r**2/2.)/(3.*m0**2+3*m0*r+r**2))
  if rotate: # rotate one of the layers
    gs = [] # empty list with geometries
    ii = 0
    for i in rot: # loop
        g = gf[ii] # get the geometry
        g = twisted_supercell(g,m0=m0,r=r) # get the twisted supercell
        print(i)
        if i!=0 and i!=1: raise # nope
        gs.append(rotate_layer(g,i*theta,dr=[0.,0.,dz*(ii-len(rot)/2.+.5)]))
        ii += 1 # increase counter
  else: raise
#  g.r = np.concatenate([g1.r,g.r,g2.r]).copy()
  g.r = np.concatenate([gi.r for gi in gs]).copy() # all the positions
#  g.r = np.concatenate([g2.r,g.r]).copy()
 # g.r = g1.r
  g.r2xyz() # update
  g.real2fractional() # update fractional coordinates 
  g = sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.])) # rotate
  g.get_fractional() # get fractional coordinates 
  return g






def rotate_layer(g,theta,dr=[0.,0.,0.]):
    """Rotate one of the layers, retainign the same unit cell"""
    g1 = g.copy() # copy geometry
    g1 = g1.rotate(theta*180/np.pi)
    g1s = g1.supercell(2) # supercell
#    g1.a1 = g.a1.copy()
#    g1.a2 = g.a2.copy()
    inuc = sculpt.sites_in_unit_cell(g1s.r,g.a1,g.a2,g.a3,dim=2)
    g1.r = g1s.r[inuc==1.]
    if g1.has_sublattice:
        g1.sublattice = g1s.sublattice[inuc==1.]
#    g1.r = sculpt.retain_unit_cell(g1s.r,g.a1,g.a2,g.a3,dim=2) # get positions
    g1.r2xyz() # update
    g1.x += dr[0] # shift upper layer
    g1.y += dr[1] # shift upper layer
    g1.z += dr[2] # shift upper layer
    g1.xyz2r() # update
    return g1 # return the geometry

def multilayer_graphene(**kwargs):
    """Return a multilayer graphene geometry"""
    g = geometry.honeycomb_lattice() # get a honeycomb lattice
    dr = g.r[0] - g.r[1] # distance
    return multilayer(g,dr=dr,dz=[0.,0.,3.],**kwargs)


def multilayer(g,l=[0],dr=[1.,0.,0.],dz=[0.,0.,1.]):
    """Return a multilayer geometry"""
    dr = np.array(dr) # to array
    dz = np.array(dz) # to array
    ss = [] # list for the sublattice
    rs = [] # list for the positions
    ii = 0 # start
    for il in l: # loop over layers
        if g.has_sublattice:
            for (ri,si) in zip(g.r,g.sublattice): # loop over positions
                rs.append(ri + dr*il + ii*dz) # add position
                ss.append(si*(-1)**(il+ii)) # add sublattice
        else:
            for ri in g.r: # loop over positions
                rs.append(ri + dr*il + ii*dz) # add position
        ii += 1
    go = g.copy()
    go.r = np.array(rs)
    if g.has_sublattice: go.sublattice = np.array(ss)
    go.r2xyz()
    go.center()
    go = sculpt.rotate_a2b(go,go.a1,np.array([1.,0.,0.]))
    return go



def mismatched_lattice(n1=5,n2=4,g=None):
    if g is None: g = geometry.honeycomb_lattice()
    g1 = g.supercell(n1) # graphene
    g2 = g.supercell(n2) # BN
    g1.z += 1.5 # shift
    g2.z -= 1.5 # shift
    g1.xyz2r() # update
    g2.xyz2r() # update
    
    gn = g1.copy() # new geometry
    if g.has_sublattice: 
        gn.sublattice = np.concatenate([g1.sublattice,g2.sublattice])
    scale = n1/n2 # scale
    g2.x *= scale # scale
    g2.y *= scale # scale
    g2.xyz2r() # update
    gn.r = np.concatenate([g1.r,g2.r]) # concatenate
    gn.r2xyz() # update
    gn.get_fractional() # update fractional coordinates
    return gn # get geometry



def twisted_multimultilayer(m0=3,
  r=1,rot=[1,0],
  g=None,dz=3.0):
  """Return the geometry for twisted multimultilayer graphene"""
  if g is None: # nothing provided, assume twisted bilayer
    g = [] # empty list
    for i in rot:
      gi = geometry.honeycomb_lattice() # default is honeycomb
      g.append(gi) # store geometry
  g0 = [] # empt list
  for gi in g:
      gi = geometry.non_orthogonal_supercell(gi,m=[[-1,0,0],[0,1,0],[0,0,1]])
      g0.append(gi) # store
  g = g0 # overwrite
  # define the rotation angle
  theta = np.arccos((3.*m0**2+3*m0*r+r**2/2.)/(3.*m0**2+3*m0*r+r**2))
  print("Theta",theta*180.0/np.pi)
  # supercell used
  nsuper = [[m0,m0+r,0],[-m0-r,2*m0+r,0],[0,0,1]]
  gs = [] # empty list
  if len(rot)!=len(g): raise # inconsistency
  zshift= 0.0 # initial shift
  for (irot,gi) in zip(rot,g): # loop
    dzi = np.max(gi.z)-np.min(gi.z) # width of this layer
    gi = geometry.non_orthogonal_supercell(gi,m=nsuper,
           reducef=lambda x: 3*np.sqrt(x))
    gi.r[:,2] -= np.min(gi.r[:,2]) # lowest layer in zero
    gi.r2xyz() # update
    if irot!=0 and irot!=1: raise # nope
    gs.append(rotate_layer(gi,irot*theta,dr=[0.,0.,zshift]))
    zshift = zshift + dzi + dz
  # now create the final geometry object
  del g # remove that list
  g = gs[0].copy() # overwrite g
  g.r = np.concatenate([gi.r for gi in gs]).copy() # all the positions
  g.sublattice = np.concatenate([gi.sublattice for gi in gs]).copy() # all the positions
  g.r2xyz() # update
  g.real2fractional() # update fractional coordinates 
  g = sculpt.rotate_a2b(g,g.a1,np.array([1.,0.,0.])) # rotate
  g.get_fractional() # get fractional coordinates 
  g.data["angle_degrees"] = theta*180.0/np.pi # store the angle of the geometry
  return g



def parse_twisted_multimultilayer(name,n=3):
    """Given a name for a multilayer, return the geometry"""
    # name is two element list
    # second element is the list with aligned stacking
    # thirs element determines whether if they are rotated
    gs = [] # empty list
    for ni in name[0]: # loop over names of the layers
            if ni in ["","A"]: gi = multilayer_graphene(l=[0]) # normal honeycomb
            elif ni=="AA": gi = multilayer_graphene(l=[0,0]) 
            elif ni=="AB": gi = multilayer_graphene(l=[0,1]) 
            elif ni=="BA": gi = multilayer_graphene(l=[1,0]) 
            elif ni=="ABC": gi = multilayer_graphene(l=[0,1,2]) 
            elif ni=="ABAB": gi = multilayer_graphene(l=[0,1,0,1]) 
            elif ni=="ABBA": gi = multilayer_graphene(l=[0,1,1,0]) 
            else: raise
            gs.append(gi) # store
    rot = name[1] # rotations
    return twisted_multimultilayer(rot=rot,g=gs,m0=n)



def multilayer_codes(n=3):
    """Return all the codes for aligned multilayers
    up to n layers"""
    out = []
    import itertools 
    for i in range(n): # loop over layer numbers
        o = list(itertools.product('ABC', repeat=i))
        o = ["A"+"".join(io) for io in o]
        out += o
    return out








from __future__ import print_function
import numpy as np
from scipy.sparse import csc_matrix

def equal(m1,m2,tol=1e-4):
  """Check if two matrices are the same"""
  if np.max(np.abs(m1-m2))>tol:
#    print(csc_matrix(m1-m2))
    print("Maximum difference",np.max(np.abs(m1-m2)))
#    print("\n")
#    print(csc_matrix(m2))
    return False
  else: return True

def check_hamiltonian(h,tol=1e-5):
  """Do various checks in Hamiltonian, to ensure that nothing weird happens"""
  hk = h.get_hk_gen() # get generator
  m = hk(np.random.random(3)) # random k-point
  if not equal(m,np.conjugate(m).T):
    print("CHECK FAILED, Hamiltonian is not Hermitian")
    print(np.round(m,2))
    exit()
#    raise # not hermitian
  if h.has_eh: # if it has electron hole degree of freedom
    v = np.random.random(3) # random kpoint
    m1 = hk(v) # Hamiltonian
    m2 = hk(-v) # Hamiltonian in time reversal point
    from .superconductivity import eh_operator
    eh = eh_operator(m1) # get the function
    if not equal(m1,-eh(m2),tol=tol): 
      print("CHECK FAILED, Hamiltonian does not have electron-hole symmetry")
      exit()
    print("CHECKED that the Hamiltonian has electron-hole symmetry")


def check_dict(mf):
    """Check a dictionary with hopping, like the one used for mean field"""
    for key in mf:
        key2 = tuple([-i for i in key])
        m1 = mf[key]
        m2 = mf[key2]
        if not equal(m1,np.conjugate(m2).T):
            print(key,key2)
            print("First")
            print(m1)
            print("Second")
            print(m2)





import numpy as np
import scipy.linalg as lg


# check whether if the fortran routines will be used
try:
  from . import tailsf90
  use_fortran = True
except:
  use_fortran=False
#  print("WARNING, Fortran functions not working in tails")



def matrix_tails(m,discard=None):
  """Return the tails of a 1d Hamiltonian"""
  es,ws = lg.eigh(m)
  ws = np.abs(np.transpose(ws)) # wavefunctions
  if callable(discard): # check if the wave is accepted
    wout = []
    eout = []
    for (e,w) in zip(es,ws):
      if discard(w): 
        wout.append(w) # store wave
        eout.append(e) # store wave
    ws = np.array(wout) # store
    es = np.array(eout) # store
  ls = np.array([loclength([w]) for w in ws]) # localization length
  np.savetxt("TAILS.OUT",np.matrix([es,ls]).T)
  return (es,ls) # return data
    



def tails(vs,use_fortran=use_fortran):
  """Return the log of the tails, centered around the maximum"""
  if use_fortran: return tailsf90.tails(vs) # fortran function
  else: return tails_python(vs) # python function



def logavgtails(vs,use_fortran=use_fortran):
  """Return the log of the average value of the tails"""
  out = tails(vs,use_fortran=use_fortran) # return all the tails
  ds = np.mean(out,axis=0) # average over waves
  ds = ds[ds>1e-10] # only sizaable elements
  return np.log(ds)



def loclength(vs,use_fortran=use_fortran):
  """Calculate the inverse localization length by fitting the
  decay of the wavefunctions"""
  out = logavgtails(vs,use_fortran=use_fortran) # get the logrho
  ns = np.array(range(len(out))) # length
  ps = np.polyfit(ns,out,1)
  return -ps[0]







def tails_python(vs,use_fortran=use_fortran):
  """Python implementation, Return the log of the tails, 
  centered around the maximum"""
  out = [] # empty list
  for v in vs: # loop over wavefunctions
    d = v*np.conjugate(v) # density
    n = len(d) # size of the array
    nmax = np.argmax(v) # index with the maximum
    ds = np.zeros(n//2) # initialize logd
    for j in range(n//2): # loop over components
      ir = (nmax + j)%n # right index
      il = (nmax - j)%n # left index
      ds[j] = d[ir] + d[il] # average tails
    out.append(ds) # store
  return out # return array


def test_tails(vs):
  """Perform a test of tails comparing Python and fortran"""
  import time
  t0 = time.perf_counter() # timing
  out1 = tails(vs,use_fortran=True) # fortran routine
  t1 = time.perf_counter() # timing
  out2 = tails(vs,use_fortran=False) # python routine
  t2 = time.perf_counter() # timing
  diff = np.mean(np.abs(np.array(out1)-np.array(out2))) # deviation
  avg = np.mean(np.abs(out1) + np.abs(out2))/2. # average
#  print("Deviation",diff)
#  print("Average",avg)
  print("Percentual error",diff/avg)
  print("Time with FORTRAN",t1-t0)
  print("Time with PYTHON",t2-t1)




from __future__ import print_function
import scipy.linalg as lg
import numpy as np
from .. import algebra


def explicit_rkky(h,ri=None,rj=None,nk=10,dj=1e-1):
    """Compute the RKKY interaction explicitly by adding
    local exchange fields"""
    if ri is None: raise
    if rj is None: raise
    h0 = h.copy() # copy Hamiltonian
    h0.turn_spinful() # spinfull Hamiltonian
    filling = h0.get_filling(nk=nk) # get the filling of the Hamiltonian
    hfe = h0.copy() # first Hamiltonian
    haf = h0.copy() # second Hamiltonian
    from ..potentials import delta
    v1 = delta(ri,v0=np.array([0.,0.,1.])*dj) # local exchange
    v2 = delta(rj,v0=np.array([0.,0.,1.])*dj) # local exchange
    v3 = delta(rj,v0=np.array([0.,0.,-1.])*dj) # local exchange
    hfe.add_exchange(v1) # add exchange
    hfe.add_exchange(v2) # add exchange
    haf.add_exchange(v1) # add exchange
    haf.add_exchange(v3) # add exchange
    # correction due to a potential shift of Fermi energy
    fermife = hfe.get_fermi4filling(filling,nk=nk) # FE
    fermiaf = haf.get_fermi4filling(filling,nk=nk) # FE
#    hfe.shift_fermi(-fermife)
#    haf.shift_fermi(-fermiaf)
    efe = hfe.get_total_energy(nk=nk,fermi=fermife) # total energy
    eaf = haf.get_total_energy(nk=nk,fermi=fermiaf) # total energy
#    print("Fermi shift",fermife-fermiaf)
    de = efe - eaf #+ (-fermife+fermiaf) # energy difference
    return -1./4.*de/dj**2 # return RKKY


# Poor man RKKY routines, using brute force energetics
# mainly for 0d, yet it can apply to 2d with PBC

def rkky_atom(hin,delta=0.001,i=None,filling=0.5):  
  if h.dimensionality != 0: raise # only for 0d
  if i is None: raise # default value
  else: r0 = hin.geometry.r[i]
  no = hin.intra.shape[0]
  # unperturbed hamiltonian
  (evals,evecs) = algebra.eigh(hin.intra)
  den0 = evecs[0]*0.0 # initializa density
  for i in range(no/2): # loop over sites
    den0 += evecs[i]*np.conjugate(evecs[i]) # density
  def denshift(delta2):
    nfill = int(round(no*filling)) # filling
    def fermi(r):
      dr = r0 - np.array(r)
      dr = dr.dot(dr)
      if dr<0.1: return delta2
  #    if x<0.0 and y>0.0: return delta
      else: return 0.0
    h = hin.copy()
    # now perturb the hamiltonian
    h.shift_fermi(fermi)
    intra = h.intra # intraterm
    (evals,evecs) = algebra.eigh(intra)
    evecs = evecs.transpose()
    eper = 0.0 # perturbed energy
    den = evecs[0]*0.0 # initializa density
    for i in range(nfill): # loop over sites
      den += evecs[i]*np.conjugate(evecs[i]) # density
    return den.real
  denup,dendn = denshift(-delta),denshift(delta) # calculate density up and dn
  den = denup - dendn
  return den # return row of correlation matrix




def rkky0d(h,check=False):
  """Calculate RKKY interaction for a 0d system"""
  if h.dimensionality != 0: raise # only for 0d
  if h.has_spin: raise # only for spinless
  if h.has_eh: raise # only for electrons
  m = [] #empty matrix
  for i in range(h.intra.shape[0]): # loop over sites
    row = rkky_atom_v1(h,i=i)
    m.append(row)
  m = np.array(m)
  diff = np.sum(np.abs(m - np.transpose(m)))
  if check: # check that it is symmetric
    if diff>0.01: raise
    print("RKKY matrix is symmetric")
  m = (m + np.transpose(m))/2.0  # symmetrize
  return m # return correlation matrix




def rkky_atom_v1(hin,delta=0.001,i=None,filling=0.5):  
  if hin.dimensionality != 0: raise # only for 0d
  if i is None: raise # default value
  else: r0 = hin.geometry.r[i]
  no = hin.intra.shape[0]
  def denshift(delta2):
    def fermi(r):
      dr = r0 - np.array(r)
      dr = dr.dot(dr)
      if dr<0.1: return delta2
  #    if x<0.0 and y>0.0: return delta
      else: return 0.0
    h = hin.copy()
    # now perturb the hamiltonian
    h.shift_fermi(fermi)
    intra = h.intra # intraterm
    (evals,evecs) = algebra.eigh(intra)
    evecs = evecs.transpose()
    return evals,evecs
  # get eigenvalues and eigenvectors for up/down
  eup,vup = denshift(-delta)
  edn,vdn = denshift(delta)
  # get fermi
  nfill = int(round(2*no*filling)) # filling
  es = sorted(np.concatenate((eup,edn))) # list of energies
  fermi = (es[nfill] + es[nfill-1])/2.0 # fermi energy
  mag = (vup[0]*0.).real # initialize
  for i in range(len(eup)):
    if eup[i]<fermi:
      dup = vup[i]*np.conjugate(vup[i]) # density up
      mag += dup.real # magnetization
    if edn[i]<fermi:
      ddn = vdn[i]*np.conjugate(vdn[i]) # density down
      mag -= ddn.real # magnetization
  return mag/delta # return row of correlation matrix



def write_rkky(h,output_file="RKKY.OUT",write_all=True,check=False):
  """ Writes in file the RKKY interaction"""
  m = rkky0d(h,check=check) # get the matrix
  evals,evecs = lg.eigh(m) # eigenvalues and eigenvectors
  evecs = np.transpose(evecs) # transpose eigenvectors
#  den = (evecs[-1]*np.conjugate(evecs[-1])).real
  den = evecs[-1].real
  ns = h.intra.shape[0] # number of sites
  fo = open(output_file,"w") # open file
  for i in range(ns):
    fo.write(str(h.geometry.x[i])+"  ")
    fo.write(str(h.geometry.y[i])+"  ")
    fo.write(str(den[i])+"\n")
  fo.close()
  if write_all:
    for j in range(len(evals)):
      name = "EIGEN_RKKY_"+str(len(evals)-j-1)+"_value_"+str(evals[j])+".OUT" # name of the file
      fo = open(name,"w")
      den = evecs[j].real
      for i in range(ns): # loop over atoms
        fo.write(str(h.geometry.x[i])+"  ")
        fo.write(str(h.geometry.y[i])+"  ")
        fo.write(str(den[i])+"\n")
      fo.close()
  # now write eigenvalues of RKKY
  fo = open("EIGENVALUES_RKKY.OUT","w")
  for i in range(len(evals)):
    fo.write(str(i)+"  ")
    fo.write(str(evals[i])+"\n")
  fo.close()
  return den




from __future__ import print_function

def intra_super2d(h,n=1,central=None):
  """ Calculates the intra of a 2d system"""
  from scipy.sparse import bmat
  from scipy.sparse import csc_matrix as csc
  tx = csc(h.tx)
  ty = csc(h.ty)
  txy = csc(h.txy)
  txmy = csc(h.txmy)
  intra = csc(h.intra)
  for i in range(n):
    intrasuper[i][i] = intra # intracell
    (x1,y1) = inds[i]
    for j in range(n):
      (x2,y2) = inds[j]
      dx = x2-x1
      dy = y2-y1
      if dx==1 and  dy==0: intrasuper[i][j] = tx
      if dx==-1 and dy==0: intrasuper[i][j] = tx.H
      if dx==0 and  dy==1: intrasuper[i][j] = ty
      if dx==0 and  dy==-1: intrasuper[i][j] = ty.H
      if dx==1 and  dy==1: intrasuper[i][j] = txy
      if dx==-1 and dy==-1: intrasuper[i][j] = txy.H
      if dx==1 and  dy==-1: intrasuper[i][j] = txmy
      if dx==-1 and dy==1: intrasuper[i][j] = txmy.H
  # substitute the central cell if it is the case
  if central!=None:
    ic = (n-1)/2
    intrasuper[ic][ic] = central
  # now convert to matrix
  intrasuper = bmat(intrasuper).todense() # supercell
  return intrasuper


def supercell1d(h,nsuper,sparse=False):
  """ Get a supercell of the system,
      h is the hamiltonian
      nsuper is the number of replicas """
  from scipy.sparse import csc_matrix as csc
  from scipy.sparse import bmat
  from copy import deepcopy
  hout = deepcopy(h) # output hamiltonian
  intra = csc(h.intra)  # intracell matrix
  inter = csc(h.inter)  # intercell matrix
  # crease supercells block matrices
  intra_super = [[None for i in range(nsuper)] for j in range(nsuper)]
  inter_super = [[None for i in range(nsuper)] for j in range(nsuper)]
  zero = csc(0.0*intra)
  for i in range(nsuper):  # onsite part
    intra_super[i][i] = intra
    inter_super[i][i] = zero
  for i in range(nsuper-1):  # inter minicell part
    intra_super[i][i+1] = inter
    intra_super[i+1][i] = inter.H
  inter_super[nsuper-1][0] = inter # inter supercell part
  inter_super[0][nsuper-1] = inter.H*0. # inter supercell part
  # create dense matrices
  if sparse:
    intra_super = bmat(intra_super)
    inter_super = csc(bmat(inter_super))
  else:
    intra_super = bmat(intra_super).todense()
    inter_super = bmat(inter_super).todense()
  # add to the output hamiltonian
  del intra
  del inter
  hout.intra = intra_super
  hout.inter = inter_super
  try: hout.geometry = h.geometry.supercell(nsuper)
  except: print("No geometry given")
  return hout






def non_orthogonal_supercell(g,m):
  """Generate a non orthogonal supercell based on a tranformation
  matrix of the unit vectors, pretty much as VESTA does"""
  a1,a2,a3 = g.a1,g.a2,g.a3 # cell vectors
  go = g.copy() # output unit cell
  # new cell vectors
  go.a1 = m[0][0]*a1 + m[0][1]*a2 + + m[0][2]*a3
  go.a2 = m[1][0]*a1 + m[1][1]*a2 + + m[1][2]*a3
  go.a3 = m[2][0]*a1 + m[2][1]*a2 + + m[2][2]*a3
  # calculate old and new volume
  vold = a1.dot(np.cross(a2,a3))  
  vnew = go.a1.dot(np.cross(go.a2,go.a3))  
  c = vnew/vold
  print("Volume of the unit cell increased by",c)
  # now create replicas until there as c times as many atoms in the
  # unit cell









import numpy as np

info = False

def get_sublattice(rs):
  """Return indexes of the sublattice, assuming that there is sublattice"""
  n = len(rs) # loop over positions
  sublattice = [0 for i in range(n)] # list for the sublattice
  ii = np.random.randint(n)
  sublattice[ii] = -1 # start with this atom
  if info: print("Looking for a sublattice")
  while True: # infinite loop
    for i in range(n): # look for a neighbor for site i
      if sublattice[i]!=0: continue
      for j in range(n): # loop over the rest of the atoms
        if sublattice[j]==0: continue # next one
        dr = rs[i] - rs[j] # distance to site i
        if 0.9<dr.dot(dr)<1.01: # if NN and sublattice 
          sublattice[i] = -sublattice[j] + 0 # opposite
          continue # next one
    if np.min(np.abs(sublattice))!=0: break
  return sublattice





import numpy as np

def radial_decay(v0=2.0,rl=3.0,voo=1.0,mode="exp"):
    """Create a fucntion that implements a radial decay
    at a certain distance"""
    if mode=="exp":
        def fs(r):
            r2 = r.dot(r)
            return voo + (v0-voo)*np.exp(-(1./rl)*np.sqrt(r2))
        return fs
    elif mode=="linear":
        def fs(r):
            r0 = np.sqrt(r.dot(r))
            if r0>rl: return voo
            else: return v0 + (voo-v0)*r0/rl
        return fs
    else: raise


